import gzip
from typing import Protocol

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import Response

import phoenix.trace.v1 as pb
from phoenix.trace.schemas import Span
from phoenix.trace.span_json_decoder import json_to_span
from phoenix.trace.v1.utils import encode


class SupportsPutSpan(Protocol):
    def put(self, span: pb.Span) -> None:
        ...


class SpanHandler(HTTPEndpoint):
    queue: SupportsPutSpan

    async def post(self, request: Request) -> Response:
        try:
            content_type = request.headers.get("content-type")
            if content_type == "application/x-protobuf":
                body = await request.body()
                content_encoding = request.headers.get("content-encoding")
                if content_encoding == "gzip":
                    body = gzip.decompress(body)
                pb_span = pb.Span()
                pb_span.ParseFromString(body)
            else:
                span = json_to_span(await request.json())
                assert isinstance(span, Span)
                pb_span = encode(span)
        except Exception:
            return Response(status_code=422)
        self.queue.put(pb_span)
        return Response()



import json
from datetime import datetime
from enum import Enum
from inspect import signature
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
)

from typing_extensions import TypeGuard

from phoenix.trace.schemas import (
    SpanAttributes,
    SpanEvent,
    SpanException,
    SpanKind,
    SpanStatusCode,
)
from phoenix.trace.semantic_conventions import (
    INPUT_MIME_TYPE,
    INPUT_VALUE,
    LLM_FUNCTION_CALL,
    LLM_INPUT_MESSAGES,
    LLM_INVOCATION_PARAMETERS,
    LLM_OUTPUT_MESSAGES,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    MESSAGE_CONTENT,
    MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON,
    MESSAGE_FUNCTION_CALL_NAME,
    MESSAGE_NAME,
    MESSAGE_ROLE,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    MimeType,
)
from phoenix.trace.utils import get_stacktrace, import_package

from ..tracer import Tracer

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletion


Parameters = Mapping[str, Any]
OpenInferenceMessage = Dict[str, str]

INSTRUMENTED_ATTRIBUTE_NAME = "is_instrumented_with_openinference_tracer"


class RequestType(Enum):
    CHAT_COMPLETION = "chat_completion"
    COMPLETION = "completion"
    EMBEDDING = "embedding"


class OpenAIInstrumentor:
    def __init__(self, tracer: Optional[Tracer] = None) -> None:
        """Instruments your OpenAI client to automatically create spans for each API call.

        Args:
            tracer (Optional[Tracer], optional): A tracer to record and handle spans. If not
            provided, the default tracer will be used.
        """
        self._tracer = tracer or Tracer()

    def instrument(self) -> None:
        """
        Instruments your OpenAI client.
        """
        openai = import_package("openai")
        is_instrumented = hasattr(
            openai.OpenAI,
            INSTRUMENTED_ATTRIBUTE_NAME,
        )
        if not is_instrumented:
            openai.OpenAI.request = _wrapped_openai_client_request_function(
                openai.OpenAI.request, self._tracer
            )
            setattr(
                openai.OpenAI,
                INSTRUMENTED_ATTRIBUTE_NAME,
                True,
            )


def _wrapped_openai_client_request_function(
    request_fn: Callable[..., Any], tracer: Tracer
) -> Callable[..., Any]:
    """Wraps the OpenAI APIRequestor.request method to create spans for each API call.

    Args:
        request_fn (Callable[..., Any]): The request method on openai.api_requestor.APIRequestor.
        tracer (Tracer): The tracer to use to create spans.

    Returns:
        Callable[..., Any]: The wrapped request method.
    """

    def wrapped(*args: Any, **kwargs: Any) -> Any:
        call_signature = signature(request_fn)
        bound_arguments = call_signature.bind(*args, **kwargs)
        is_streaming = bound_arguments.arguments["stream"]
        options = bound_arguments.arguments["options"]
        parameters = options.json_data
        url = options.url
        current_status_code = SpanStatusCode.UNSET
        events: List[SpanEvent] = []
        attributes: SpanAttributes = dict()
        if not is_streaming and _get_request_type(url) is RequestType.CHAT_COMPLETION:
            for (
                attribute_name,
                get_parameter_attribute_fn,
            ) in _PARAMETER_ATTRIBUTE_FUNCTIONS.items():
                if (attribute_value := get_parameter_attribute_fn(parameters)) is not None:
                    attributes[attribute_name] = attribute_value
            response = None
            try:
                start_time = datetime.now()
                response = request_fn(*args, **kwargs)
                end_time = datetime.now()
                current_status_code = SpanStatusCode.OK
                return response
            except Exception as error:
                end_time = datetime.now()
                current_status_code = SpanStatusCode.ERROR
                events.append(
                    SpanException(
                        message=str(error),
                        timestamp=end_time,
                        exception_type=type(error).__name__,
                        exception_stacktrace=get_stacktrace(error),
                    )
                )
                raise
            finally:
                if _is_chat_completion(response):
                    for (
                        attribute_name,
                        get_chat_completion_attribute_fn,
                    ) in _CHAT_COMPLETION_ATTRIBUTE_FUNCTIONS.items():
                        if (
                            attribute_value := get_chat_completion_attribute_fn(response)
                        ) is not None:
                            attributes[attribute_name] = attribute_value
                tracer.create_span(
                    name="OpenAI Chat Completion",
                    span_kind=SpanKind.LLM,
                    start_time=start_time,
                    end_time=end_time,
                    status_code=current_status_code,
                    status_message="",
                    attributes=attributes,
                    events=events,
                )
        else:
            return request_fn(*args, **kwargs)

    return wrapped


def _input_value(parameters: Parameters) -> str:
    return json.dumps(parameters)


def _input_mime_type(_: Any) -> MimeType:
    return MimeType.JSON


def _llm_input_messages(parameters: Parameters) -> Optional[List[OpenInferenceMessage]]:
    if not (messages := parameters.get("messages")):
        return None
    return [_to_openinference_message(message, expects_name=True) for message in messages]


def _llm_invocation_parameters(
    parameters: Parameters,
) -> str:
    return json.dumps(parameters)


def _output_value(chat_completion: "ChatCompletion") -> str:
    return chat_completion.json()


def _output_mime_type(_: Any) -> MimeType:
    return MimeType.JSON


def _llm_output_messages(chat_completion: "ChatCompletion") -> List[OpenInferenceMessage]:
    return [
        _to_openinference_message(choice.message.dict(), expects_name=False)
        for choice in chat_completion.choices
    ]


def _llm_token_count_prompt(chat_completion: "ChatCompletion") -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.prompt_tokens
    return None


def _llm_token_count_completion(chat_completion: "ChatCompletion") -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.completion_tokens
    return None


def _llm_token_count_total(chat_completion: "ChatCompletion") -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.total_tokens
    return None


def _llm_function_call(
    chat_completion: "ChatCompletion",
) -> Optional[str]:
    choices = chat_completion.choices
    choice = choices[0]
    if choice.finish_reason == "function_call" and (function_call := choice.message.function_call):
        return function_call.json()
    return None


def _get_request_type(url: str) -> Optional[RequestType]:
    """Get OpenAI request type from URL, or returns None if the request type cannot be recognized"""
    if "chat/completions" in url:
        return RequestType.CHAT_COMPLETION
    if "completions" in url:
        return RequestType.COMPLETION
    if "embeddings" in url:
        return RequestType.EMBEDDING
    return None


def _to_openinference_message(
    message: Mapping[str, Any], *, expects_name: bool
) -> OpenInferenceMessage:
    """Converts an OpenAI input or output message to an OpenInference message.

    Args:
        message (Dict[str, Any]): The OpenAI message to be parsed.

        expects_name (bool): Whether to parse the "name" key in the OpenAI message. This key is
        sometimes included in "function"-role input messages to specify the function name, but is
        not included in output messages.

    Returns:
        OpenInferenceMessage: A message in OpenInference format.
    """
    openinference_message = {}
    if role := message.get("role"):
        openinference_message[MESSAGE_ROLE] = role
    if content := message.get("content"):
        openinference_message[MESSAGE_CONTENT] = content
    if function_call_data := message.get("function_call"):
        if function_name := function_call_data.get("name"):
            openinference_message[MESSAGE_FUNCTION_CALL_NAME] = function_name
        if function_arguments := function_call_data.get("arguments"):
            openinference_message[MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON] = function_arguments
    if expects_name and (name := message.get("name")):
        openinference_message[MESSAGE_NAME] = name
    return openinference_message


_PARAMETER_ATTRIBUTE_FUNCTIONS: Dict[str, Callable[[Parameters], Any]] = {
    INPUT_VALUE: _input_value,
    INPUT_MIME_TYPE: _input_mime_type,
    LLM_INPUT_MESSAGES: _llm_input_messages,
    LLM_INVOCATION_PARAMETERS: _llm_invocation_parameters,
}
_CHAT_COMPLETION_ATTRIBUTE_FUNCTIONS: Dict[str, Callable[["ChatCompletion"], Any]] = {
    OUTPUT_VALUE: _output_value,
    OUTPUT_MIME_TYPE: _output_mime_type,
    LLM_OUTPUT_MESSAGES: _llm_output_messages,
    LLM_TOKEN_COUNT_PROMPT: _llm_token_count_prompt,
    LLM_TOKEN_COUNT_COMPLETION: _llm_token_count_completion,
    LLM_TOKEN_COUNT_TOTAL: _llm_token_count_total,
    LLM_FUNCTION_CALL: _llm_function_call,
}


def _is_chat_completion(response: Any) -> TypeGuard["ChatCompletion"]:
    """
    Type guard for ChatCompletion.
    """
    openai = import_package("openai")
    return isinstance(response, openai.types.chat.ChatCompletion)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: trace/v1/trace.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14trace/v1/trace.proto\x12\x16phoenix.proto.trace.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1egoogle/protobuf/wrappers.proto\"\xa2\x0b\n\x04Span\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x35\n\x07\x63ontext\x18\x02 \x01(\x0b\x32$.phoenix.proto.trace.v1.Span.Context\x12\x33\n\x0eparent_span_id\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.BytesValue\x12\x0c\n\x04name\x18\x04 \x01(\t\x12.\n\nstart_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x33\n\x06status\x18\x07 \x01(\x0b\x32#.phoenix.proto.trace.v1.Span.Status\x12\x32\n\x06\x65vents\x18\x08 \x03(\x0b\x32\".phoenix.proto.trace.v1.Span.Event\x12:\n\nexceptions\x18\t \x03(\x0b\x32&.phoenix.proto.trace.v1.Span.Exception\x12\x33\n\x05input\x18\n \x01(\x0b\x32$.phoenix.proto.trace.v1.Span.IOValue\x12\x34\n\x06output\x18\x0b \x01(\x0b\x32$.phoenix.proto.trace.v1.Span.IOValue\x12\x0c\n\x04kind\x18\x0c \x01(\t\x12\x34\n\tretrieval\x18\r \x01(\x0b\x32!.phoenix.proto.trace.v1.Retrieval\x12\x34\n\tembedding\x18\x0e \x01(\x0b\x32!.phoenix.proto.trace.v1.Embedding\x12(\n\x03llm\x18\x0f \x01(\x0b\x32\x1b.phoenix.proto.trace.v1.LLM\x12*\n\x04tool\x18\x10 \x01(\x0b\x32\x1c.phoenix.proto.trace.v1.Tool\x1a,\n\x07\x43ontext\x12\x10\n\x08trace_id\x18\x01 \x01(\x0c\x12\x0f\n\x07span_id\x18\x02 \x01(\x0c\x1a\x95\x01\n\x06Status\x12\x36\n\x04\x63ode\x18\x01 \x01(\x0e\x32(.phoenix.proto.trace.v1.Span.Status.Code\x12-\n\x07message\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\"$\n\x04\x43ode\x12\t\n\x05UNSET\x10\x00\x12\x06\n\x02OK\x10\x01\x12\t\n\x05\x45RROR\x10\x02\x1aq\n\x05\x45vent\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0c\n\x04name\x18\x02 \x01(\t\x12-\n\ttimestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\xa1\x02\n\tException\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12-\n\x07message\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12*\n\x04type\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12+\n\x07\x65scaped\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x30\n\nstacktrace\x18\x05 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12-\n\ttimestamp\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1aX\n\x07IOValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12-\n\njson_value\x18\x02 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x42\x06\n\x04kind\"\xdf\x02\n\tRetrieval\x12=\n\tdocuments\x18\x01 \x03(\x0b\x32*.phoenix.proto.trace.v1.Retrieval.Document\x1a\x92\x02\n\x08\x44ocument\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12(\n\x02id\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12*\n\x05score\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12-\n\x07\x63ontent\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12)\n\x08metadata\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12)\n\x04hash\x18\x06 \x01(\x0b\x32\x1b.google.protobuf.BytesValue\"\xf4\x01\n\tEmbedding\x12?\n\nembeddings\x18\x01 \x03(\x0b\x32+.phoenix.proto.trace.v1.Embedding.Embedding\x12\x30\n\nmodel_name\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x1at\n\tEmbedding\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0e\n\x06vector\x18\x02 \x03(\x02\x12*\n\x04text\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\"2\n\x03LLM\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"3\n\x04Tool\x12+\n\nattributes\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Structb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'trace.v1.trace_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SPAN._serialized_start=144
  _SPAN._serialized_end=1586
  _SPAN_CONTEXT._serialized_start=893
  _SPAN_CONTEXT._serialized_end=937
  _SPAN_STATUS._serialized_start=940
  _SPAN_STATUS._serialized_end=1089
  _SPAN_STATUS_CODE._serialized_start=1053
  _SPAN_STATUS_CODE._serialized_end=1089
  _SPAN_EVENT._serialized_start=1091
  _SPAN_EVENT._serialized_end=1204
  _SPAN_EXCEPTION._serialized_start=1207
  _SPAN_EXCEPTION._serialized_end=1496
  _SPAN_IOVALUE._serialized_start=1498
  _SPAN_IOVALUE._serialized_end=1586
  _RETRIEVAL._serialized_start=1589
  _RETRIEVAL._serialized_end=1940
  _RETRIEVAL_DOCUMENT._serialized_start=1666
  _RETRIEVAL_DOCUMENT._serialized_end=1940
  _EMBEDDING._serialized_start=1943
  _EMBEDDING._serialized_end=2187
  _EMBEDDING_EMBEDDING._serialized_start=2071
  _EMBEDDING_EMBEDDING._serialized_end=2187
  _LLM._serialized_start=2189
  _LLM._serialized_end=2239
  _TOOL._serialized_start=2241
  _TOOL._serialized_end=2292
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from typing import List, Optional, cast
from urllib import request

from phoenix.trace.trace_dataset import TraceDataset
from phoenix.trace.utils import json_lines_to_df


@dataclass(frozen=True)
class TracesFixture:
    name: str
    description: str
    file_name: str


llama_index_rag_fixture = TracesFixture(
    name="llama_index_rag",
    description="Traces from running the llama_index on a RAG use case.",
    file_name="llama_index_rag_with_rerank.jsonl",
)

llama_index_calculator_agent_fixture = TracesFixture(
    name="llama_index_calculator_agent",
    description="Traces from running the llama_index with calculator tools.",
    file_name="llama_index_calculator_agent_v2.jsonl",
)

llama_index_rag_fixture_with_davinci = TracesFixture(
    name="llama_index_rag_with_davinci",
    description="Traces from running llama_index on a RAG use case with the completions API.",
    file_name="llama_index_rag_with_davinci_v0.jsonl",
)

langchain_rag_stuff_document_chain_fixture = TracesFixture(
    name="langchain_rag_stuff_document_chain",
    description="LangChain RAG data",
    file_name="langchain_rag.jsonl",
)

langchain_titanic_csv_agent_evaluator_fixture = TracesFixture(
    name="lc_titanic",
    description="LangChain titanic.csv Agent Evaluator",
    file_name="lc_titanic.jsonl",
)

langchain_qa_with_sources_fixture = TracesFixture(
    name="langchain_qa_with_sources",
    description="LangChain QA with sources on financial data",
    file_name="langchain_qa_with_sources_chain.jsonl",
)

random_fixture = TracesFixture(
    name="random",
    description="Randomly generated traces",
    file_name="random.jsonl",
)

TRACES_FIXTURES: List[TracesFixture] = [
    llama_index_rag_fixture,
    llama_index_rag_fixture_with_davinci,
    langchain_rag_stuff_document_chain_fixture,
    langchain_titanic_csv_agent_evaluator_fixture,
    random_fixture,
    langchain_qa_with_sources_fixture,
    llama_index_calculator_agent_fixture,
]

NAME_TO_TRACES_FIXTURE = {fixture.name: fixture for fixture in TRACES_FIXTURES}


def _get_trace_fixture_by_name(fixture_name: str) -> TracesFixture:
    """
    Returns the fixture whose name matches the input name.

    Raises
    ------
    ValueError
        if the input fixture name does not match any known fixture names.
    """
    if fixture_name not in NAME_TO_TRACES_FIXTURE:
        valid_fixture_names = ", ".join(NAME_TO_TRACES_FIXTURE.keys())
        raise ValueError(f'"{fixture_name}" is invalid. Valid names are: {valid_fixture_names}')
    return NAME_TO_TRACES_FIXTURE[fixture_name]


def _download_traces_fixture(
    fixture: TracesFixture,
    host: Optional[str] = "https://storage.googleapis.com/",
    bucket: Optional[str] = "arize-assets",
    prefix: Optional[str] = "phoenix/traces/",
) -> List[str]:
    """
    Downloads the traces fixture from the phoenix bucket.
    """
    url = f"{host}{bucket}/{prefix}{fixture.file_name}"
    with request.urlopen(url) as f:
        return cast(List[str], f.readlines())


def load_example_traces(use_case: str) -> TraceDataset:
    """
    Loads a trace dataframe by name.

    NB: this functionality is under active construction.
    """
    fixture = _get_trace_fixture_by_name(use_case)
    return TraceDataset(json_lines_to_df(_download_traces_fixture(fixture)))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import google.protobuf.timestamp_pb2
import google.protobuf.wrappers_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Span(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Context(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TRACE_ID_FIELD_NUMBER: builtins.int
        SPAN_ID_FIELD_NUMBER: builtins.int
        trace_id: builtins.bytes
        span_id: builtins.bytes
        def __init__(
            self,
            *,
            trace_id: builtins.bytes = ...,
            span_id: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["span_id", b"span_id", "trace_id", b"trace_id"]) -> None: ...

    @typing_extensions.final
    class Status(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _Code:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _CodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Span.Status._Code.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            UNSET: Span.Status._Code.ValueType  # 0
            OK: Span.Status._Code.ValueType  # 1
            ERROR: Span.Status._Code.ValueType  # 2

        class Code(_Code, metaclass=_CodeEnumTypeWrapper): ...
        UNSET: Span.Status.Code.ValueType  # 0
        OK: Span.Status.Code.ValueType  # 1
        ERROR: Span.Status.Code.ValueType  # 2

        CODE_FIELD_NUMBER: builtins.int
        MESSAGE_FIELD_NUMBER: builtins.int
        code: global___Span.Status.Code.ValueType
        @property
        def message(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        def __init__(
            self,
            *,
            code: global___Span.Status.Code.ValueType = ...,
            message: google.protobuf.wrappers_pb2.StringValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["message", b"message"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["code", b"code", "message", b"message"]) -> None: ...

    @typing_extensions.final
    class Event(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ATTRIBUTES_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        TIMESTAMP_FIELD_NUMBER: builtins.int
        @property
        def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
        name: builtins.str
        @property
        def timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        def __init__(
            self,
            *,
            attributes: google.protobuf.struct_pb2.Struct | None = ...,
            name: builtins.str = ...,
            timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "timestamp", b"timestamp"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "name", b"name", "timestamp", b"timestamp"]) -> None: ...

    @typing_extensions.final
    class Exception(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ATTRIBUTES_FIELD_NUMBER: builtins.int
        MESSAGE_FIELD_NUMBER: builtins.int
        TYPE_FIELD_NUMBER: builtins.int
        ESCAPED_FIELD_NUMBER: builtins.int
        STACKTRACE_FIELD_NUMBER: builtins.int
        TIMESTAMP_FIELD_NUMBER: builtins.int
        @property
        def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
        @property
        def message(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        @property
        def type(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        @property
        def escaped(self) -> google.protobuf.wrappers_pb2.BoolValue: ...
        @property
        def stacktrace(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        @property
        def timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        def __init__(
            self,
            *,
            attributes: google.protobuf.struct_pb2.Struct | None = ...,
            message: google.protobuf.wrappers_pb2.StringValue | None = ...,
            type: google.protobuf.wrappers_pb2.StringValue | None = ...,
            escaped: google.protobuf.wrappers_pb2.BoolValue | None = ...,
            stacktrace: google.protobuf.wrappers_pb2.StringValue | None = ...,
            timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "escaped", b"escaped", "message", b"message", "stacktrace", b"stacktrace", "timestamp", b"timestamp", "type", b"type"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "escaped", b"escaped", "message", b"message", "stacktrace", b"stacktrace", "timestamp", b"timestamp", "type", b"type"]) -> None: ...

    @typing_extensions.final
    class IOValue(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STRING_VALUE_FIELD_NUMBER: builtins.int
        JSON_VALUE_FIELD_NUMBER: builtins.int
        string_value: builtins.str
        @property
        def json_value(self) -> google.protobuf.struct_pb2.Struct: ...
        def __init__(
            self,
            *,
            string_value: builtins.str = ...,
            json_value: google.protobuf.struct_pb2.Struct | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["json_value", b"json_value", "kind", b"kind", "string_value", b"string_value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["json_value", b"json_value", "kind", b"kind", "string_value", b"string_value"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal["kind", b"kind"]) -> typing_extensions.Literal["string_value", "json_value"] | None: ...

    ATTRIBUTES_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    PARENT_SPAN_ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    START_TIME_FIELD_NUMBER: builtins.int
    END_TIME_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    EVENTS_FIELD_NUMBER: builtins.int
    EXCEPTIONS_FIELD_NUMBER: builtins.int
    INPUT_FIELD_NUMBER: builtins.int
    OUTPUT_FIELD_NUMBER: builtins.int
    KIND_FIELD_NUMBER: builtins.int
    RETRIEVAL_FIELD_NUMBER: builtins.int
    EMBEDDING_FIELD_NUMBER: builtins.int
    LLM_FIELD_NUMBER: builtins.int
    TOOL_FIELD_NUMBER: builtins.int
    @property
    def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
    @property
    def context(self) -> global___Span.Context: ...
    @property
    def parent_span_id(self) -> google.protobuf.wrappers_pb2.BytesValue: ...
    name: builtins.str
    @property
    def start_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def end_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def status(self) -> global___Span.Status: ...
    @property
    def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Span.Event]: ...
    @property
    def exceptions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Span.Exception]: ...
    @property
    def input(self) -> global___Span.IOValue: ...
    @property
    def output(self) -> global___Span.IOValue: ...
    kind: builtins.str
    @property
    def retrieval(self) -> global___Retrieval: ...
    @property
    def embedding(self) -> global___Embedding: ...
    @property
    def llm(self) -> global___LLM: ...
    @property
    def tool(self) -> global___Tool: ...
    def __init__(
        self,
        *,
        attributes: google.protobuf.struct_pb2.Struct | None = ...,
        context: global___Span.Context | None = ...,
        parent_span_id: google.protobuf.wrappers_pb2.BytesValue | None = ...,
        name: builtins.str = ...,
        start_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        end_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        status: global___Span.Status | None = ...,
        events: collections.abc.Iterable[global___Span.Event] | None = ...,
        exceptions: collections.abc.Iterable[global___Span.Exception] | None = ...,
        input: global___Span.IOValue | None = ...,
        output: global___Span.IOValue | None = ...,
        kind: builtins.str = ...,
        retrieval: global___Retrieval | None = ...,
        embedding: global___Embedding | None = ...,
        llm: global___LLM | None = ...,
        tool: global___Tool | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "context", b"context", "embedding", b"embedding", "end_time", b"end_time", "input", b"input", "llm", b"llm", "output", b"output", "parent_span_id", b"parent_span_id", "retrieval", b"retrieval", "start_time", b"start_time", "status", b"status", "tool", b"tool"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "context", b"context", "embedding", b"embedding", "end_time", b"end_time", "events", b"events", "exceptions", b"exceptions", "input", b"input", "kind", b"kind", "llm", b"llm", "name", b"name", "output", b"output", "parent_span_id", b"parent_span_id", "retrieval", b"retrieval", "start_time", b"start_time", "status", b"status", "tool", b"tool"]) -> None: ...

global___Span = Span

@typing_extensions.final
class Retrieval(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Document(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ATTRIBUTES_FIELD_NUMBER: builtins.int
        ID_FIELD_NUMBER: builtins.int
        SCORE_FIELD_NUMBER: builtins.int
        CONTENT_FIELD_NUMBER: builtins.int
        METADATA_FIELD_NUMBER: builtins.int
        HASH_FIELD_NUMBER: builtins.int
        @property
        def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
        @property
        def id(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        @property
        def score(self) -> google.protobuf.wrappers_pb2.FloatValue: ...
        @property
        def content(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        @property
        def metadata(self) -> google.protobuf.struct_pb2.Struct: ...
        @property
        def hash(self) -> google.protobuf.wrappers_pb2.BytesValue: ...
        def __init__(
            self,
            *,
            attributes: google.protobuf.struct_pb2.Struct | None = ...,
            id: google.protobuf.wrappers_pb2.StringValue | None = ...,
            score: google.protobuf.wrappers_pb2.FloatValue | None = ...,
            content: google.protobuf.wrappers_pb2.StringValue | None = ...,
            metadata: google.protobuf.struct_pb2.Struct | None = ...,
            hash: google.protobuf.wrappers_pb2.BytesValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "content", b"content", "hash", b"hash", "id", b"id", "metadata", b"metadata", "score", b"score"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "content", b"content", "hash", b"hash", "id", b"id", "metadata", b"metadata", "score", b"score"]) -> None: ...

    DOCUMENTS_FIELD_NUMBER: builtins.int
    @property
    def documents(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Retrieval.Document]: ...
    def __init__(
        self,
        *,
        documents: collections.abc.Iterable[global___Retrieval.Document] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["documents", b"documents"]) -> None: ...

global___Retrieval = Retrieval

@typing_extensions.final
class Embedding(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Embedding(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ATTRIBUTES_FIELD_NUMBER: builtins.int
        VECTOR_FIELD_NUMBER: builtins.int
        TEXT_FIELD_NUMBER: builtins.int
        @property
        def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
        @property
        def vector(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
        @property
        def text(self) -> google.protobuf.wrappers_pb2.StringValue: ...
        def __init__(
            self,
            *,
            attributes: google.protobuf.struct_pb2.Struct | None = ...,
            vector: collections.abc.Iterable[builtins.float] | None = ...,
            text: google.protobuf.wrappers_pb2.StringValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "text", b"text"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "text", b"text", "vector", b"vector"]) -> None: ...

    EMBEDDINGS_FIELD_NUMBER: builtins.int
    MODEL_NAME_FIELD_NUMBER: builtins.int
    @property
    def embeddings(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Embedding.Embedding]: ...
    @property
    def model_name(self) -> google.protobuf.wrappers_pb2.StringValue: ...
    def __init__(
        self,
        *,
        embeddings: collections.abc.Iterable[global___Embedding.Embedding] | None = ...,
        model_name: google.protobuf.wrappers_pb2.StringValue | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["model_name", b"model_name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["embeddings", b"embeddings", "model_name", b"model_name"]) -> None: ...

global___Embedding = Embedding

@typing_extensions.final
class LLM(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTES_FIELD_NUMBER: builtins.int
    @property
    def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
    def __init__(
        self,
        *,
        attributes: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes"]) -> None: ...

global___LLM = LLM

@typing_extensions.final
class Tool(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATTRIBUTES_FIELD_NUMBER: builtins.int
    @property
    def attributes(self) -> google.protobuf.struct_pb2.Struct: ...
    def __init__(
        self,
        *,
        attributes: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["attributes", b"attributes"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes"]) -> None: ...

global___Tool = Tool



import json
from datetime import datetime, timezone
from itertools import chain
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    SupportsFloat,
    Tuple,
    Union,
    cast,
)
from uuid import UUID

from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue, BytesValue, FloatValue, StringValue

import phoenix.trace.v1 as pb
from phoenix.trace.schemas import (
    Span,
    SpanContext,
    SpanEvent,
    SpanException,
    SpanKind,
    SpanStatusCode,
)
from phoenix.trace.semantic_conventions import (
    DOCUMENT_CONTENT,
    DOCUMENT_ID,
    DOCUMENT_METADATA,
    DOCUMENT_SCORE,
    EMBEDDING_EMBEDDINGS,
    EMBEDDING_MODEL_NAME,
    EMBEDDING_TEXT,
    EMBEDDING_VECTOR,
    EXCEPTION_ESCAPED,
    EXCEPTION_MESSAGE,
    EXCEPTION_STACKTRACE,
    EXCEPTION_TYPE,
    INPUT_MIME_TYPE,
    INPUT_VALUE,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    RETRIEVAL_DOCUMENTS,
    MimeType,
)


def encode(span: Span) -> pb.Span:
    _attributes: Mapping[str, Any] = dict(span.attributes)
    retrieval, _attributes = _excise_retrieval(_attributes)
    embedding, _attributes = _excise_embedding(_attributes)
    input, _attributes = _excise_input(_attributes)
    output, _attributes = _excise_output(_attributes)
    status = pb.Span.Status(message=_maybe_str(span.status_message))
    if span.status_code is SpanStatusCode.ERROR:
        status.code = pb.Span.Status.Code.ERROR
    elif span.status_code is SpanStatusCode.OK:
        status.code = pb.Span.Status.Code.OK
    parent_span_id = BytesValue(value=span.parent_id.bytes) if span.parent_id else None
    pb_span = pb.Span(
        start_time=_as_timestamp(span.start_time),
        end_time=_maybe_timestamp(span.end_time),
        status=status,
        name=span.name,
        kind=span.span_kind.value,
        context=pb.Span.Context(
            trace_id=span.context.trace_id.bytes,
            span_id=span.context.span_id.bytes,
        ),
        parent_span_id=parent_span_id,
        attributes=_maybe_struct(_attributes),
        input=input,
        output=output,
        retrieval=retrieval,
        embedding=embedding,
    )
    for event in span.events:
        if event.name == "exception":
            pb_span.exceptions.append(
                _encode_exception(
                    cast(SpanException, event),
                ),
            )
        else:
            pb_span.events.append(
                _encode_event(event),
            )
    return pb_span


def decode(
    pb_span: pb.Span,
) -> Span:
    trace_id = UUID(bytes=pb_span.context.trace_id)
    span_id = UUID(bytes=pb_span.context.span_id)
    parent_id = (
        UUID(bytes=pb_span.parent_span_id.value) if pb_span.HasField("parent_span_id") else None
    )
    start_time = pb_span.start_time.ToDatetime(timezone.utc)
    end_time = pb_span.end_time.ToDatetime(timezone.utc) if pb_span.HasField("end_time") else None
    attributes = MessageToDict(pb_span.attributes)
    if pb_span.HasField("input"):
        attributes.update(_decode_input(pb_span.input))
    if pb_span.HasField("output"):
        attributes.update(_decode_output(pb_span.output))
    if pb_span.HasField("retrieval"):
        attributes.update(_decode_retrieval(pb_span.retrieval))
    if pb_span.HasField("embedding"):
        attributes.update(_decode_embedding(pb_span.embedding))
    events = sorted(
        chain(
            map(_decode_event, pb_span.events),
            map(_decode_exception, pb_span.exceptions),
        ),
        key=lambda event: event.timestamp,
    )
    status_code = SpanStatusCode.UNSET
    if pb_span.status.code == pb.Span.Status.Code.OK:
        status_code = SpanStatusCode.OK
    elif pb_span.status.code == pb.Span.Status.Code.ERROR:
        status_code = SpanStatusCode.ERROR
    return Span(
        name=pb_span.name,
        context=SpanContext(trace_id=trace_id, span_id=span_id),
        parent_id=parent_id,
        span_kind=SpanKind(pb_span.kind.upper()),
        start_time=start_time,
        end_time=end_time,
        attributes=attributes,
        status_code=status_code,
        status_message=pb_span.status.message.value,
        conversation=None,
        events=events,
    )


def _excise_input(
    attributes: Mapping[str, Any],
) -> Tuple[Optional[pb.Span.IOValue], Dict[str, Any]]:
    _attributes = dict(attributes)
    if not (
        _attributes.keys()
        & {
            INPUT_VALUE,
            INPUT_MIME_TYPE,
        }
    ):
        return None, _attributes
    input_value: Optional[str] = _attributes.pop(INPUT_VALUE, None)
    assert input_value is None or isinstance(
        input_value, str
    ), f"{INPUT_VALUE} must be str, found {type(input_value)}"
    input_mime_type: Optional[MimeType] = _attributes.pop(INPUT_MIME_TYPE, None)
    assert input_mime_type is None or isinstance(
        input_mime_type, MimeType
    ), f"{INPUT_MIME_TYPE} must be MimeType, found {type(input_mime_type)}"
    return (
        _encode_io_value(
            input_value or "",
            input_mime_type,
        ),
        _attributes,
    )


def _excise_output(
    attributes: Mapping[str, Any],
) -> Tuple[Optional[pb.Span.IOValue], Dict[str, Any]]:
    _attributes = dict(attributes)
    if not (
        _attributes.keys()
        & {
            OUTPUT_VALUE,
            OUTPUT_MIME_TYPE,
        }
    ):
        return None, _attributes
    output_value: Optional[str] = _attributes.pop(OUTPUT_VALUE, None)
    assert output_value is None or isinstance(
        output_value, str
    ), f"{OUTPUT_VALUE} must be str, found {type(output_value)}"
    output_mime_type: Optional[MimeType] = _attributes.pop(OUTPUT_MIME_TYPE, None)
    assert output_mime_type is None or isinstance(
        output_mime_type, MimeType
    ), f"{OUTPUT_MIME_TYPE} must be MimeType, found {type(output_mime_type)}"
    return (
        _encode_io_value(
            output_value or "",
            output_mime_type,
        ),
        _attributes,
    )


def _excise_retrieval(
    attributes: Mapping[str, Any],
) -> Tuple[Optional[pb.Retrieval], Dict[str, Any]]:
    _attributes = dict(attributes)
    if not (
        _attributes.keys()
        & {
            RETRIEVAL_DOCUMENTS,
        }
    ):
        return None, _attributes
    documents: Optional[Iterable[Mapping[str, Any]]] = _attributes.pop(RETRIEVAL_DOCUMENTS, None)
    assert documents is None or isinstance(
        documents, Iterable
    ), f"{RETRIEVAL_DOCUMENTS} must be Iterable, found {type(documents)}"
    return (
        _encode_retrieval(documents=() if documents is None else documents),
        _attributes,
    )


def _excise_embedding(
    attributes: Mapping[str, Any],
) -> Tuple[Optional[pb.Embedding], Dict[str, Any]]:
    _attributes = dict(attributes)
    if not (
        _attributes.keys()
        & {
            EMBEDDING_EMBEDDINGS,
            EMBEDDING_MODEL_NAME,
        }
    ):
        return None, _attributes
    embeddings: Optional[Iterable[Mapping[str, Any]]] = _attributes.pop(EMBEDDING_EMBEDDINGS, None)
    assert embeddings is None or isinstance(
        embeddings, Iterable
    ), f"{EMBEDDING_EMBEDDINGS} must be Mapping, found {type(embeddings)}"
    model_name: Optional[str] = _attributes.pop(EMBEDDING_MODEL_NAME, None)
    assert model_name is None or isinstance(
        model_name, str
    ), f"{EMBEDDING_MODEL_NAME} must be str, found {type(model_name)}"
    return (
        _encode_embedding(
            embeddings=embeddings or (),
            model_name=model_name,
        ),
        _attributes,
    )


def _encode_event(
    span_event: SpanEvent,
) -> pb.Span.Event:
    timestamp = Timestamp()
    timestamp.FromDatetime(span_event.timestamp)
    attributes = None
    if span_event.attributes:
        attributes = Struct()
        attributes.update(span_event.attributes)
    pb_span_event = pb.Span.Event(
        name=span_event.name,
        timestamp=timestamp,
        attributes=attributes,
    )
    return pb_span_event


def _decode_event(
    pb_span_event: pb.Span.Event,
) -> SpanEvent:
    return SpanEvent(
        name=pb_span_event.name,
        timestamp=pb_span_event.timestamp.ToDatetime(timezone.utc),
        attributes=MessageToDict(pb_span_event.attributes),
    )


def _encode_exception(
    span_exception: SpanException,
) -> pb.Span.Exception:
    timestamp = Timestamp()
    timestamp.FromDatetime(span_exception.timestamp)
    _attributes: Dict[str, Any] = dict(span_exception.attributes)
    exception_message: Optional[str] = _attributes.pop(EXCEPTION_MESSAGE, None)
    assert exception_message is None or isinstance(
        exception_message, str
    ), f"{EXCEPTION_MESSAGE} must be str, found {type(exception_message)}"
    exception_type: Optional[str] = _attributes.pop(EXCEPTION_TYPE, None)
    assert exception_type is None or isinstance(
        exception_type, str
    ), f"{EXCEPTION_TYPE} must be str, found {type(exception_type)}"
    exception_escaped: Optional[bool] = _attributes.pop(EXCEPTION_ESCAPED, None)
    assert exception_escaped is None or isinstance(
        exception_escaped, bool
    ), f"{EXCEPTION_ESCAPED} must be bool, found {type(exception_escaped)}"
    exception_stacktrace: Optional[str] = _attributes.pop(EXCEPTION_STACKTRACE, None)
    assert exception_stacktrace is None or isinstance(
        exception_stacktrace, str
    ), f"{EXCEPTION_STACKTRACE} must be str, found {type(exception_stacktrace)}"
    pb_span_exception = pb.Span.Exception(
        timestamp=timestamp,
        message=_maybe_str(exception_message),
        type=_maybe_str(exception_type),
        escaped=_maybe_bool(exception_escaped),
        stacktrace=_maybe_str(exception_stacktrace),
        attributes=_maybe_struct(_attributes),
    )
    return pb_span_exception


def _decode_exception(
    pb_span_exception: pb.Span.Exception,
) -> SpanException:
    exception_message = (
        pb_span_exception.message.value if pb_span_exception.HasField("message") else None
    )
    exception_type = pb_span_exception.type.value if pb_span_exception.HasField("type") else None
    exception_escaped = (
        pb_span_exception.escaped.value if pb_span_exception.HasField("escaped") else None
    )
    exception_stacktrace = (
        pb_span_exception.stacktrace.value if pb_span_exception.HasField("stacktrace") else None
    )
    span_exception = SpanException(
        timestamp=pb_span_exception.timestamp.ToDatetime(timezone.utc),
        message=exception_message or "",
        exception_type=exception_type,
        exception_escaped=exception_escaped,
        exception_stacktrace=exception_stacktrace,
    )
    span_exception.attributes.update(
        MessageToDict(
            pb_span_exception.attributes,
        ),
    )
    return span_exception


def _decode_input(
    pb_io_value: pb.Span.IOValue,
) -> Iterator[Tuple[str, Union[str, MimeType]]]:
    return zip(
        (INPUT_VALUE, INPUT_MIME_TYPE),
        _decode_io_value(pb_io_value),
    )


def _decode_output(
    pb_io_value: pb.Span.IOValue,
) -> Iterator[Tuple[str, Union[str, MimeType]]]:
    return zip(
        (OUTPUT_VALUE, OUTPUT_MIME_TYPE),
        _decode_io_value(pb_io_value),
    )


def _encode_io_value(
    io_value: str,
    mime_type: Optional[MimeType],
) -> pb.Span.IOValue:
    if mime_type is MimeType.JSON:
        struct = Struct()
        if io_value:
            struct.update(json.loads(io_value))
        return pb.Span.IOValue(json_value=struct)
    return pb.Span.IOValue(string_value=io_value)


def _decode_io_value(
    pb_io_value: pb.Span.IOValue,
) -> Iterator[Union[str, MimeType]]:
    if pb_io_value.WhichOneof("kind") == "json_value":
        yield json.dumps(MessageToDict(pb_io_value.json_value))
        yield MimeType.JSON
    else:
        yield pb_io_value.string_value


def _encode_retrieval(
    documents: Iterable[Mapping[str, Any]],
) -> pb.Retrieval:
    return pb.Retrieval(
        documents=map(
            _encode_retrieval_document,
            documents,
        ),
    )


def _decode_retrieval(
    pb_retrieval: pb.Retrieval,
) -> Iterator[Tuple[str, Any]]:
    yield (
        RETRIEVAL_DOCUMENTS,
        [
            dict(_decode_retrieval_document(pb_retrieval_document))
            for pb_retrieval_document in pb_retrieval.documents
        ],
    )


def _encode_retrieval_document(
    document: Mapping[str, Any],
) -> pb.Retrieval.Document:
    _attributes: Dict[str, Any] = dict(document)
    document_id: Optional[str] = _attributes.pop(DOCUMENT_ID, None)
    assert document_id is None or isinstance(
        document_id, str
    ), f"{DOCUMENT_ID} must be str, found {type(document_id)}"
    document_score: Optional[float] = _attributes.pop(DOCUMENT_SCORE, None)
    assert document_score is None or isinstance(
        document_score, SupportsFloat
    ), f"{DOCUMENT_SCORE} must be float, found {type(document_score)}"
    document_content: Optional[str] = _attributes.pop(DOCUMENT_CONTENT, None)
    assert document_content is None or isinstance(
        document_content, str
    ), f"{DOCUMENT_CONTENT} must be str, found {type(document_content)}"
    document_metadata: Optional[Mapping[str, Any]] = _attributes.pop(DOCUMENT_METADATA, None)
    assert document_metadata is None or isinstance(
        document_metadata, Mapping
    ), f"{DOCUMENT_METADATA} must be Mapping, found {type(document_metadata)}"
    return pb.Retrieval.Document(
        id=_maybe_str(document_id),
        score=_maybe_float(document_score),
        content=_maybe_str(document_content),
        metadata=_maybe_struct(document_metadata),
        attributes=_maybe_struct(_attributes),
    )


def _decode_retrieval_document(
    pb_document: pb.Retrieval.Document,
) -> Iterator[Tuple[str, Any]]:
    if pb_document.HasField("id"):
        yield DOCUMENT_ID, pb_document.id.value
    if pb_document.HasField("score"):
        yield DOCUMENT_SCORE, pb_document.score.value
    if pb_document.HasField("content"):
        yield DOCUMENT_CONTENT, pb_document.content.value
    if pb_document.HasField("metadata"):
        yield DOCUMENT_METADATA, MessageToDict(pb_document.metadata)
    if pb_document.HasField("attributes"):
        yield from MessageToDict(pb_document.attributes).items()


def _encode_embedding(
    embeddings: Iterable[Mapping[str, Any]],
    model_name: Optional[str],
) -> pb.Embedding:
    return pb.Embedding(
        model_name=_maybe_str(model_name),
        embeddings=map(_encode_embedding_embedding, embeddings),
    )


def _decode_embedding(
    pb_embedding: pb.Embedding,
) -> Iterator[Tuple[str, Any]]:
    if pb_embedding.HasField("model_name"):
        yield EMBEDDING_MODEL_NAME, pb_embedding.model_name.value
    yield (
        EMBEDDING_EMBEDDINGS,
        [
            dict(_decode_embedding_embedding(pb_embedding_embedding))
            for pb_embedding_embedding in pb_embedding.embeddings
        ],
    )


def _encode_embedding_embedding(
    embedding: Mapping[str, Any],
) -> pb.Embedding.Embedding:
    _attributes = dict(embedding)
    vector: Optional[Iterable[float]] = _attributes.pop(EMBEDDING_VECTOR, None)
    assert vector is None or isinstance(
        vector, Iterable
    ), f"{EMBEDDING_VECTOR} must be Iterable, found {type(vector)}"
    embedding_text: Optional[str] = _attributes.pop(EMBEDDING_TEXT, None)
    assert embedding_text is None or isinstance(
        embedding_text, str
    ), f"{EMBEDDING_TEXT} must be str, found {type(embedding_text)}"
    return pb.Embedding.Embedding(
        vector=() if vector is None else vector,
        text=_maybe_str(embedding_text),
        attributes=_maybe_struct(_attributes),
    )


def _decode_embedding_embedding(
    pb_embedding_embedding: pb.Embedding.Embedding,
) -> Iterator[Tuple[str, Any]]:
    yield EMBEDDING_VECTOR, list(pb_embedding_embedding.vector)
    if pb_embedding_embedding.HasField("text"):
        yield EMBEDDING_TEXT, pb_embedding_embedding.text.value
    if pb_embedding_embedding.HasField("attributes"):
        yield from MessageToDict(pb_embedding_embedding.attributes).items()


def _maybe_str(obj: Optional[str]) -> Optional[StringValue]:
    return None if not obj else StringValue(value=obj)


def _maybe_float(obj: Optional[float]) -> Optional[FloatValue]:
    return None if obj is None else FloatValue(value=obj)


def _maybe_bool(obj: Optional[bool]) -> Optional[BoolValue]:
    return None if obj is None else BoolValue(value=obj)


def _as_timestamp(obj: datetime) -> Timestamp:
    timestamp = Timestamp()
    timestamp.FromDatetime(obj)
    return timestamp


def _maybe_timestamp(obj: Optional[datetime]) -> Optional[Timestamp]:
    return _as_timestamp(obj) if obj else None


def _as_struct(obj: Mapping[str, Any]) -> Struct:
    struct = Struct()
    struct.update(obj)
    return struct


def _maybe_struct(obj: Optional[Mapping[str, Any]]) -> Optional[Struct]:
    return _as_struct(obj) if obj else None



import logging
from typing import Any, Dict, List, Optional

from llama_index.callbacks.base_handler import BaseCallbackHandler
from llama_index.callbacks.schema import CBEventType

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

CBEventID = str


class DebugCallbackHandler(BaseCallbackHandler):
    def _print_event(self, payload: Dict[Any, Any]) -> None:
        for k, v in payload.items():
            print(f"**{k}: **\n{v}")
            print("*" * 50)

    def __init__(self) -> None:
        super().__init__(event_starts_to_ignore=[], event_ends_to_ignore=[])

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: CBEventID = "",
        parent_id: CBEventID = "",
        **kwargs: Any,
    ) -> CBEventID:
        return event_id

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: CBEventID = "",
        **kwargs: Any,
    ) -> None:
        if payload is not None:
            self._print_event(payload)

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return



from enum import Enum
from functools import partial
from typing import Any, Iterable, Iterator

import pandas as pd
import strawberry

from phoenix.core.traces import (
    END_TIME,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    START_TIME,
    ComputedAttributes,
)
from phoenix.server.api.types.SortDir import SortDir
from phoenix.trace.schemas import Span


@strawberry.enum
class SpanColumn(Enum):
    startTime = START_TIME
    endTime = END_TIME
    latencyMs = ComputedAttributes.LATENCY_MS.value
    tokenCountTotal = LLM_TOKEN_COUNT_TOTAL
    tokenCountPrompt = LLM_TOKEN_COUNT_PROMPT
    tokenCountCompletion = LLM_TOKEN_COUNT_COMPLETION
    cumulativeTokenCountTotal = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_TOTAL.value
    cumulativeTokenCountPrompt = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_PROMPT.value
    cumulativeTokenCountCompletion = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_COMPLETION.value


@strawberry.input
class SpanSort:
    """
    The sort column and direction for span connections
    """

    col: SpanColumn
    dir: SortDir

    def __call__(self, spans: Iterable[Span]) -> Iterator[Span]:
        """
        Sorts the spans by the given column and direction
        """
        yield from pd.Series(spans, dtype=object).sort_values(
            key=lambda series: series.apply(partial(_get_column, span_column=self.col)),
            ascending=self.dir.value == SortDir.asc.value,
        )


def _get_column(span: Span, span_column: SpanColumn) -> Any:
    if span_column is SpanColumn.startTime:
        return span.start_time
    if span_column is SpanColumn.endTime:
        return span.end_time
    return span.attributes.get(span_column.value)



import json
from io import BytesIO
from urllib.error import HTTPError
from urllib.request import urlopen
from zipfile import ZipFile

import pandas as pd


def download_benchmark_dataset(task: str, dataset_name: str) -> pd.DataFrame:
    """Downloads an Arize evals benchmark dataset as a pandas dataframe.

    Args:
        task (str): Task to be performed.
        dataset_name (str): Name of the dataset.

    Returns:
        pandas.DataFrame: A pandas dataframe containing the data.
    """
    jsonl_file_name = f"{dataset_name}.jsonl"
    url = f"http://storage.googleapis.com/arize-assets/phoenix/evals/{task}/{jsonl_file_name}.zip"
    try:
        with urlopen(url) as response:
            zip_byte_stream = BytesIO(response.read())
            with ZipFile(zip_byte_stream) as zip_file:
                with zip_file.open(jsonl_file_name) as jsonl_file:
                    return pd.DataFrame(map(json.loads, jsonl_file.readlines()))
    except HTTPError:
        raise ValueError(f'Dataset "{dataset_name}" for "{task}" task does not exist.')


def get_tqdm_progress_bar_formatter(title: str) -> str:
    """Returns a progress bar formatter for use with tqdm.

    Args:
        title (str): The title of the progress bar, displayed as a prefix.

    Returns:
        str: A formatter to be passed to the bar_format argument of tqdm.
    """
    return (
        title + " |{bar}| {n_fmt}/{total_fmt} ({percentage:3.1f}%) "
        "| ⏳ {elapsed}<{remaining} | {rate_fmt}{postfix}"
    )



from typing import List, Optional

from phoenix.experimental.evals import PromptTemplate
from phoenix.experimental.evals.models import BaseEvalModel


class MapReducer:
    """
    Evaluates data that is too large to fit into a single context window using a
    map-reduce strategy. The data must first be divided into "chunks" that
    individually fit into an LLM's context window. Each chunk of data is
    individually evaluated (the "map" step), producing intermediate outputs that
    are combined into a single result (the "reduce" step).

    This is the simplest strategy for evaluating long-context data.
    """

    def __init__(
        self,
        model: BaseEvalModel,
        map_prompt_template: PromptTemplate,
        reduce_prompt_template: PromptTemplate,
    ) -> None:
        """Initializes an instance.

        Args:
            model (BaseEvalModel): The LLM model to use for evaluation.

            map_prompt_template (PromptTemplate): The template that is mapped
            over each chunk to produce intermediate outputs. Must contain the
            {chunk} placeholder.

            reduce_prompt_template (PromptTemplate): The template that combines
            the intermediate outputs into a single result. Must contain the
            {mapped} placeholder, which will be formatted as a list of the
            intermediate outputs produced by the map step.
        """
        self._model = model
        self._map_prompt_template = map_prompt_template
        self._reduce_prompt_template = reduce_prompt_template

    def evaluate(self, chunks: List[str]) -> str:
        """Evaluates a list of two or more chunks.

        Args:
            chunks (List[str]): A list of chunks to be evaluated. Each chunk is
            inserted into the map_prompt_template and must therefore fit within
            the LLM's context window and still leave room for the rest of the
            prompt.

        Returns:
            str: The output of the map-reduce process.
        """
        if len(chunks) < 2:
            raise ValueError(
                "The map-reduce strategy is not needed to evaluate data "
                "that fits within a single context window. "
                "Consider using llm_classify instead."
            )
        model = self._model
        mapped_records = []
        for chunk in chunks:
            map_prompt = self._map_prompt_template.format({"chunk": chunk})
            intermediate_output = model(map_prompt)
            mapped_records.append(intermediate_output)
        reduce_prompt = self._reduce_prompt_template.format({"mapped": repr(mapped_records)})
        return model(reduce_prompt)


class Refiner:
    """
    Evaluates data that is too large to fit into a single context window using a
    refine strategy. The data must first be divided into "chunks" that
    individually fit into an LLM's context window. An initial "accumulator" is
    generated from the first chunk of data. The accumulator is subsequently
    refined by iteratively updating and incorporating new information from each
    subsequent chunk. An optional synthesis step can be used to synthesize the
    final accumulator into a desired format.
    """

    def __init__(
        self,
        model: BaseEvalModel,
        initial_prompt_template: PromptTemplate,
        refine_prompt_template: PromptTemplate,
        synthesize_prompt_template: Optional[PromptTemplate] = None,
    ) -> None:
        """Initializes an instance.

        Args:
            model (BaseEvalModel): The LLM model to use for evaluation.

            initial_prompt_template (PromptTemplate): The template for the
            initial invocation of the model that will generate the initial
            accumulator. Should contain the {chunk} placeholder.

            refine_prompt_template (PromptTemplate): The template for refining
            the accumulator across all subsequence chunks. Must contain the
            {chunk} and {accumulator} placeholders.

            synthesize_prompt_template (Optional[PromptTemplate], optional): An
            optional template to synthesize the final version of the
            accumulator. Must contain the {accumulator} placeholder.
        """
        self._model = model
        self._initial_prompt_template = initial_prompt_template
        self._refine_prompt_template = refine_prompt_template
        self._synthesize_prompt_template = synthesize_prompt_template

    def evaluate(self, chunks: List[str]) -> str:
        """Evaluates a list of two or more chunks.

        Args:
            chunks (List[str]): A list of chunks to be evaluated. Each chunk is
            inserted into the initial_prompt_template and refine_prompt_template
            and must therefore fit within the LLM's context window and still
            leave room for the rest of the prompt.

        Returns:
            str: The output of the refine process.
        """
        if len(chunks) < 2:
            raise ValueError(
                "The refine strategy is not needed to evaluate data "
                "that fits within a single context window. "
                "Consider using llm_classify instead."
            )
        model = self._model
        initial_prompt = self._initial_prompt_template.format({"chunk": chunks[0]})
        accumulator = model(initial_prompt)
        for chunk in chunks[1:]:
            refine_prompt = self._refine_prompt_template.format(
                {"accumulator": accumulator, "chunk": chunk}
            )
            accumulator = model(refine_prompt)
        if not self._synthesize_prompt_template:
            return accumulator
        reduce_prompt = self._synthesize_prompt_template.format({"accumulator": accumulator})
        return model(reduce_prompt)



import json
import uuid
from datetime import datetime
from typing import Iterator, List, Optional, cast

import pandas as pd
from pandas import DataFrame, read_parquet

from phoenix.datetime_utils import normalize_timestamps

from ..config import DATASET_DIR, GENERATED_DATASET_NAME_PREFIX
from .schemas import ATTRIBUTE_PREFIX, CONTEXT_PREFIX, Span
from .span_json_decoder import json_to_span
from .span_json_encoder import span_to_json

# A set of columns that is required
REQUIRED_COLUMNS = [
    "name",
    "span_kind",
    "parent_id",
    "start_time",
    "end_time",
    "status_code",
    "status_message",
    "context.span_id",
    "context.trace_id",
]


def normalize_dataframe(dataframe: DataFrame) -> "DataFrame":
    """Makes the dataframe have appropriate data types"""

    # Convert the start and end times to datetime
    dataframe["start_time"] = normalize_timestamps(dataframe["start_time"])
    dataframe["end_time"] = normalize_timestamps(dataframe["end_time"])
    return dataframe


class TraceDataset:
    """
    A TraceDataset is a wrapper around a dataframe which is a flattened representation
    of Spans. The collection of spans trace the LLM application's execution.

    Parameters
    __________
    dataframe: pandas.DataFrame
        the pandas dataframe containing the tracing data. Each row represents a span.
    """

    name: str
    dataframe: pd.DataFrame
    _data_file_name: str = "data.parquet"

    def __init__(self, dataframe: DataFrame, name: Optional[str] = None):
        # Validate the the dataframe has required fields
        if missing_columns := set(REQUIRED_COLUMNS) - set(dataframe.columns):
            raise ValueError(
                f"The dataframe is missing some required columns: {', '.join(missing_columns)}"
            )
        self.dataframe = normalize_dataframe(dataframe)
        self.name = name or f"{GENERATED_DATASET_NAME_PREFIX}{str(uuid.uuid4())}"

    @classmethod
    def from_spans(cls, spans: List[Span]) -> "TraceDataset":
        """Creates a TraceDataset from a list of spans.

        Args:
            spans (List[Span]): A list of spans.

        Returns:
            TraceDataset: A TraceDataset containing the spans.
        """
        return cls(
            pd.json_normalize(
                (json.loads(span_to_json(span)) for span in spans),  # type: ignore
                max_level=1,
            )
        )

    def to_spans(self) -> Iterator[Span]:
        for _, row in self.dataframe.iterrows():
            is_attribute = row.index.str.startswith(ATTRIBUTE_PREFIX)
            attribute_keys = row.index[is_attribute]
            attributes = (
                row.loc[is_attribute]
                .rename(
                    {key: key[len(ATTRIBUTE_PREFIX) :] for key in attribute_keys},
                )
                .dropna()
                .to_dict()
            )
            is_context = row.index.str.startswith(CONTEXT_PREFIX)
            context_keys = row.index[is_context]
            context = (
                row.loc[is_context]
                .rename(
                    {key: key[len(CONTEXT_PREFIX) :] for key in context_keys},
                )
                .to_dict()
            )
            end_time: Optional[datetime] = cast(datetime, row.get("end_time"))
            if end_time is pd.NaT:
                end_time = None
            yield json_to_span(
                {
                    "name": row["name"],
                    "context": context,
                    "span_kind": row["span_kind"],
                    "parent_id": row.get("parent_id"),
                    "start_time": cast(datetime, row["start_time"]).isoformat(),
                    "end_time": end_time.isoformat() if end_time else None,
                    "status_code": row["status_code"],
                    "status_message": row.get("status_message") or "",
                    "attributes": attributes,
                    "events": row.get("events") or [],
                    "conversation": row.get("conversation"),
                }
            )

    @classmethod
    def from_name(cls, name: str) -> "TraceDataset":
        """Retrieves a dataset by name from the file system"""
        directory = DATASET_DIR / name
        df = read_parquet(directory / cls._data_file_name)
        return cls(df, name)

    def to_disc(self) -> None:
        """writes the data to disc"""
        directory = DATASET_DIR / self.name
        directory.mkdir(parents=True, exist_ok=True)
        self.dataframe.to_parquet(
            directory / self._data_file_name,
            allow_truncated_timestamps=True,
            coerce_timestamps="ms",
        )



import logging
from typing import Any, Callable, Dict, Optional, Union

import pandas as pd

from phoenix.experimental.evals.models import BaseEvalModel, set_verbosity
from phoenix.experimental.evals.templates import (
    PromptTemplate,
    map_template,
    normalize_prompt_template,
)

logger = logging.getLogger(__name__)


def _no_op_parser(response: str) -> Dict[str, str]:
    return {"output": response}


def llm_generate(
    dataframe: pd.DataFrame,
    template: Union[PromptTemplate, str],
    model: BaseEvalModel,
    system_instruction: Optional[str] = None,
    verbose: bool = False,
    output_parser: Optional[Callable[[str], Dict[str, Any]]] = None,
) -> pd.DataFrame:
    """
    Generates a text using a template using an LLM. This function is useful
    if you want to generate synthetic data, such as irrelevant responses
    Args:
        dataframe (pandas.DataFrame): A pandas dataframe in which each row
        represents a record to be used as in input to the template. All
        template variable names must appear as column names in the dataframe
        (extra columns unrelated to the template are permitted).

        template (Union[PromptTemplate, str]): The prompt template as either an
        instance of PromptTemplate or a string. If the latter, the variable
        names should be surrounded by curly braces so that a call to `.format`
        can be made to substitute variable values.

        model (BaseEvalModel): An LLM model class.

        system_instruction (Optional[str], optional): An optional system
        message.

        verbose (bool, optional): If True, prints detailed information to stdout such as model
        invocation parameters and retry info. Default False.

        output_parser (Callable[[str], Dict[str, Any]], optional): An optional function
        that takes each generated response and parses it to a dictionary. The keys of the dictionary
        should correspond to the column names of the output dataframe. If None, the output dataframe
        will have a single column named "output". Default None.

    Returns:
        generations_dataframe (pandas.DataFrame): A dataframe where each row
        represents the generated output

    """
    output_parser = output_parser or _no_op_parser
    with set_verbosity(model, verbose) as verbose_model:
        template = normalize_prompt_template(template)
        logger.info(f"Template: \n{template.prompt()}\n")
        logger.info(f"Template variables: {template.variables}")
        prompts = map_template(dataframe, template)

        # For each prompt, generate and parse the response
        output = []
        # Wrap the loop in a try / catch so that we can still return a dataframe
        # even if the process is interrupted
        try:
            for prompt in prompts:
                logger.info(f"Prompt: {prompt}")
                response = verbose_model(prompt, instruction=system_instruction)
                parsed_response = output_parser(response)
                output.append(parsed_response)

        except (Exception, KeyboardInterrupt) as e:
            logger.error(e)
            print(
                "Process was interrupted. The return value will be incomplete",
                e,
            )
        # Return the data as a dataframe
        return pd.DataFrame(output)



from enum import Enum
from typing import Any, Optional

import strawberry

from phoenix.trace import semantic_conventions


@strawberry.enum
class MimeType(Enum):
    text = semantic_conventions.MimeType.TEXT
    json = semantic_conventions.MimeType.JSON

    @classmethod
    def _missing_(cls, v: Any) -> Optional["MimeType"]:
        return None if v else cls.text



import json
import logging
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import pandas as pd
from tqdm.auto import tqdm

from phoenix.experimental.evals.models import BaseEvalModel, OpenAIModel, set_verbosity
from phoenix.experimental.evals.templates import (
    NOT_PARSABLE,
    RAG_RELEVANCY_PROMPT_RAILS_MAP,
    RAG_RELEVANCY_PROMPT_TEMPLATE,
    ClassificationTemplate,
    PromptOptions,
    PromptTemplate,
    map_template,
    normalize_classification_template,
)
from phoenix.experimental.evals.utils import get_tqdm_progress_bar_formatter
from phoenix.trace.semantic_conventions import DOCUMENT_CONTENT, INPUT_VALUE, RETRIEVAL_DOCUMENTS
from phoenix.utilities.logging import printif

logger = logging.getLogger(__name__)


OPENINFERENCE_QUERY_COLUMN_NAME = "attributes." + INPUT_VALUE
OPENINFERENCE_DOCUMENT_COLUMN_NAME = "attributes." + RETRIEVAL_DOCUMENTS

# argument keys in the default openai function call,
# defined here only to prevent typos
_RESPONSE = "response"
_EXPLANATION = "explanation"


def llm_classify(
    dataframe: pd.DataFrame,
    model: BaseEvalModel,
    template: Union[ClassificationTemplate, PromptTemplate, str],
    rails: List[str],
    system_instruction: Optional[str] = None,
    verbose: bool = False,
    use_function_calling_if_available: bool = True,
    provide_explanation: bool = False,
) -> pd.DataFrame:
    """Classifies each input row of the dataframe using an LLM. Returns a pandas.DataFrame
    where the first column is named `label` and contains the classification labels. An optional
    column named `explanation` is added when `provide_explanation=True`.

    Args:
        dataframe (pandas.DataFrame): A pandas dataframe in which each row represents a record to be
        classified. All template variable names must appear as column names in the dataframe (extra
        columns unrelated to the template are permitted).

        template (Union[ClassificationTemplate, PromptTemplate, str]): The prompt template as
        either an instance of PromptTemplate, ClassificationTemplate or a string. If a string, the
        variable names should be surrounded by curly braces so that a call to `.format` can be made
        to substitute variable values.

        model (BaseEvalModel): An LLM model class.

        rails (List[str]): A list of strings representing the possible output classes of the model's
        predictions.

        system_instruction (Optional[str], optional): An optional system message.

        verbose (bool, optional): If True, prints detailed info to stdout such as model invocation
        parameters and details about retries and snapping to rails. Default False.

        use_function_calling_if_available (bool, default=True): If True, use function calling
        (if available) as a means to constrain the LLM outputs. With function calling, the LLM
        is instructed to provide its response as a structured JSON object, which is easier
        to parse.

        provide_explanation (bool, default=False): If True, provides an explanation for each
        classification label. A column named `explanation` is added to the output dataframe.
        Currently, this is only available for models with function calling.

    Returns:
        pandas.DataFrame: A dataframe where the `label` column (at column position 0) contains
        the classification labels. If provide_explanation=True, then an additional column named
        `explanation` is added to contain the explanation for each label. The dataframe has
        the same length and index as the input dataframe. The classification label values are
        from the entries in the rails argument or "NOT_PARSABLE" if the model's output could
        not be parsed.
    """
    use_openai_function_call = (
        use_function_calling_if_available
        and isinstance(model, OpenAIModel)
        and model.supports_function_calling
    )

    model_kwargs: Dict[str, Any] = {}
    if use_openai_function_call:
        openai_function = _default_openai_function(rails, provide_explanation)
        model_kwargs["functions"] = [openai_function]
        model_kwargs["function_call"] = {"name": openai_function["name"]}

    eval_template = normalize_classification_template(rails=rails, template=template)

    prompt_options = PromptOptions(provide_explanation=provide_explanation)
    prompts = map_template(dataframe, eval_template, options=prompt_options)

    labels: List[Optional[str]] = [None] * len(dataframe)
    explanations: List[Optional[str]] = [None] * len(dataframe)

    printif(verbose, f"Using prompt:\n\n{eval_template.prompt(prompt_options)}")
    if generation_info := model.verbose_generation_info():
        printif(verbose, generation_info)

    # Wrap the loop in a try / catch so that we can still return a dataframe
    # even if the process is interrupted
    try:
        for index, prompt in enumerate(
            tqdm(prompts, bar_format=get_tqdm_progress_bar_formatter("llm_classify"))
        ):
            with set_verbosity(model, verbose) as verbose_model:
                response = verbose_model(prompt, instruction=system_instruction, **model_kwargs)
            if not use_openai_function_call:
                if provide_explanation:
                    unrailed_label, explanation = (
                        eval_template.extract_label_from_explanation(response),
                        response,
                    )
                    printif(
                        verbose and unrailed_label == NOT_PARSABLE,
                        f"- Could not parse {repr(response)}",
                    )
                else:
                    unrailed_label = response
                    explanation = None
            else:
                try:
                    function_arguments = json.loads(response, strict=False)
                    unrailed_label = function_arguments.get(_RESPONSE)
                    explanation = function_arguments.get(_EXPLANATION)
                except json.JSONDecodeError:
                    unrailed_label = response
                    explanation = None
            labels[index] = _snap_to_rail(unrailed_label, rails, verbose=verbose)
            explanations[index] = explanation
    except (Exception, KeyboardInterrupt) as e:
        logger.error(e)
        print(
            "Process was interrupted. The return value will be incomplete",
            e,
        )

    return pd.DataFrame(
        data={
            "label": labels,
            **({"explanation": explanations} if provide_explanation else {}),
        },
        index=dataframe.index,
    )


def run_relevance_eval(
    dataframe: pd.DataFrame,
    model: BaseEvalModel,
    template: Union[ClassificationTemplate, str] = RAG_RELEVANCY_PROMPT_TEMPLATE,
    rails: List[str] = list(RAG_RELEVANCY_PROMPT_RAILS_MAP.values()),
    system_instruction: Optional[str] = None,
    query_column_name: str = "input",
    document_column_name: str = "reference",
    verbose: bool = False,
) -> List[List[str]]:
    """
    Given a pandas dataframe containing queries and retrieved documents, classifies the relevance of
    each retrieved document to the corresponding query using an LLM.

    Args:
        dataframe (pd.DataFrame): A pandas dataframe containing queries and retrieved documents. If
        both query_column_name and reference_column_name are present in the input dataframe, those
        columns are used as inputs and should appear in the following format:

        - The entries of the query column must be strings.
        - The entries of the documents column must be lists of strings. Each list may contain an
          arbitrary number of document texts retrieved for the corresponding query.

        If the input dataframe is lacking either query_column_name or reference_column_name but has
        query and retrieved document columns in OpenInference trace format named
        "attributes.input.value" and "attributes.retrieval.documents", respectively, then those
        columns are used as inputs and should appear in the following format:

        - The entries of the query column must be strings.
        - The entries of the document column must be lists of OpenInference document objects, each
          object being a dictionary that stores the document text under the key "document.content".

        This latter format is intended for running evaluations on exported OpenInference trace
        dataframes. For more information on the OpenInference tracing specification, see
        https://github.com/Arize-ai/open-inference-spec/.

        model (BaseEvalModel): The model used for evaluation.

        template (Union[PromptTemplate, str], optional): The template used for evaluation.

        rails (List[str], optional): A list of strings representing the possible output classes of
        the model's predictions.

        query_column_name (str, optional): The name of the query column in the dataframe, which
        should also be a template variable.

        reference_column_name (str, optional): The name of the document column in the dataframe,
        which should also be a template variable.

        system_instruction (Optional[str], optional): An optional system message.

        verbose (bool, optional): If True, prints detailed information to stdout such as model
        invocation parameters and retry info. Default False.

    Returns:
        List[List[str]]: A list of relevant and not relevant classifications. The "shape" of the
        list should mirror the "shape" of the retrieved documents column, in the sense that it has
        the same length as the input dataframe and each sub-list has the same length as the
        corresponding list in the retrieved documents column. The values in the sub-lists are either
        entries from the rails argument or "NOT_PARSABLE" in the case where the LLM output could not
        be parsed.
    """

    with set_verbosity(model, verbose) as verbose_model:
        query_column = dataframe.get(query_column_name)
        document_column = dataframe.get(document_column_name)
        if query_column is None or document_column is None:
            openinference_query_column = dataframe.get(OPENINFERENCE_QUERY_COLUMN_NAME)
            openinference_document_column = dataframe.get(OPENINFERENCE_DOCUMENT_COLUMN_NAME)
            if openinference_query_column is None or openinference_document_column is None:
                raise ValueError(
                    f'Dataframe columns must include either "{query_column_name}" and '
                    f'"{document_column_name}", or "{OPENINFERENCE_QUERY_COLUMN_NAME}" and '
                    f'"{OPENINFERENCE_DOCUMENT_COLUMN_NAME}".'
                )
            query_column = openinference_query_column
            document_column = openinference_document_column.map(
                lambda docs: _get_contents_from_openinference_documents(docs)
                if docs is not None
                else None
            )

        queries = cast("pd.Series[str]", query_column).tolist()
        document_lists = cast("pd.Series[str]", document_column).tolist()
        indexes = []
        expanded_queries = []
        expanded_documents = []
        for index, (query, documents) in enumerate(zip(queries, document_lists)):
            if query is None or documents is None:
                continue
            for document in documents:
                indexes.append(index)
                expanded_queries.append(query)
                expanded_documents.append(document)
        predictions = llm_classify(
            dataframe=pd.DataFrame(
                {
                    query_column_name: expanded_queries,
                    document_column_name: expanded_documents,
                }
            ),
            model=verbose_model,
            template=template,
            rails=rails,
            system_instruction=system_instruction,
            verbose=verbose,
        ).iloc[:, 0]
        outputs: List[List[str]] = [[] for _ in range(len(dataframe))]
        for index, prediction in zip(indexes, predictions):
            outputs[index].append(prediction)
        return outputs


def _get_contents_from_openinference_documents(documents: Iterable[Any]) -> List[Optional[str]]:
    """
    Get document contents from an iterable of OpenInference document objects, which are dictionaries
    containing the document text under the "document.content" key.
    """
    return [doc.get(DOCUMENT_CONTENT) if isinstance(doc, dict) else None for doc in documents]


def _snap_to_rail(raw_string: Optional[str], rails: List[str], verbose: bool = False) -> str:
    """
    Snaps a string to the nearest rail, or returns None if the string cannot be
    snapped to a rail.

    Args:
        raw_string (str): An input to be snapped to a rail.

        rails (List[str]): The target set of strings to snap to.

    Returns:
        str: A string from the rails argument or "UNPARSABLE" if the input
        string could not be snapped.
    """
    if not raw_string:
        return NOT_PARSABLE
    snap_string = raw_string.lower()
    rails = list(set(rail.lower() for rail in rails))
    rails.sort(key=len, reverse=True)
    found_rails = set()
    for rail in rails:
        if rail in snap_string:
            found_rails.add(rail)
            snap_string = snap_string.replace(rail, "")
    if len(found_rails) != 1:
        printif(verbose, f"- Cannot snap {repr(raw_string)} to rails")
        return NOT_PARSABLE
    rail = list(found_rails)[0]
    printif(verbose, f"- Snapped {repr(raw_string)} to rail: {rail}")
    return rail


def _default_openai_function(
    rails: List[str],
    with_explanation: bool = False,
) -> Dict[str, Any]:
    properties = {
        _RESPONSE: {"type": "string", "description": "Your response.", "enum": rails},
        **(
            {
                _EXPLANATION: {
                    "type": "string",
                    "description": "Explanation of the reasoning for your response.",
                },
            }
            if with_explanation
            else {}
        ),
    }
    required = [_RESPONSE, *([_EXPLANATION] if with_explanation else [])]
    return {
        "name": "record_response",
        "description": "A function to record your response.",
        "parameters": {
            "type": "object",
            "properties": properties,
            "required": required,
        },
    }



import json
import logging
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, Iterable, Iterator, List, Mapping, Optional, Tuple
from uuid import UUID

from langchain.callbacks.tracers.base import BaseTracer
from langchain.callbacks.tracers.schemas import Run
from langchain.load.dump import dumpd
from langchain.schema.messages import BaseMessage

from phoenix.trace.exporter import HttpExporter
from phoenix.trace.schemas import Span, SpanEvent, SpanException, SpanKind, SpanStatusCode
from phoenix.trace.semantic_conventions import (
    DOCUMENT_CONTENT,
    DOCUMENT_METADATA,
    INPUT_MIME_TYPE,
    INPUT_VALUE,
    LLM_FUNCTION_CALL,
    LLM_INPUT_MESSAGES,
    LLM_INVOCATION_PARAMETERS,
    LLM_MODEL_NAME,
    LLM_OUTPUT_MESSAGES,
    LLM_PROMPT_TEMPLATE,
    LLM_PROMPT_TEMPLATE_VARIABLES,
    LLM_PROMPT_TEMPLATE_VERSION,
    LLM_PROMPTS,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    MESSAGE_CONTENT,
    MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON,
    MESSAGE_FUNCTION_CALL_NAME,
    MESSAGE_ROLE,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    RETRIEVAL_DOCUMENTS,
    TOOL_DESCRIPTION,
    TOOL_NAME,
    MimeType,
)
from phoenix.trace.tracer import Tracer
from phoenix.utilities.error_handling import graceful_fallback

logger = logging.getLogger(__name__)


Message = Dict[str, Any]


def _langchain_run_type_to_span_kind(run_type: str) -> SpanKind:
    # TODO: LangChain is moving away from enums and to arbitrary strings
    # for the run_type variable, so we may need to do the same
    try:
        return SpanKind(run_type.upper())
    except ValueError:
        return SpanKind.UNKNOWN


def _serialize_json(obj: Any) -> str:
    if isinstance(obj, datetime):
        return obj.isoformat()
    return str(obj)


def _convert_io(obj: Optional[Dict[str, Any]]) -> Iterator[Any]:
    if not obj:
        return
    if not isinstance(obj, dict):
        raise ValueError(f"obj should be dict, but obj={obj}")
    if len(obj) == 1 and isinstance(value := next(iter(obj.values())), str):
        yield value
    else:
        yield json.dumps(obj, default=_serialize_json)
        yield MimeType.JSON


def _prompts(run_inputs: Dict[str, Any]) -> Iterator[Tuple[str, List[str]]]:
    """Yields prompts if present."""
    if "prompts" in run_inputs:
        yield LLM_PROMPTS, run_inputs["prompts"]


def _input_messages(run_inputs: Mapping[str, Any]) -> Iterator[Tuple[str, List[Message]]]:
    """Yields chat messages if present."""
    if not hasattr(run_inputs, "get"):
        return
    # There may be more than one set of messages. We'll use just the first set.
    if not (multiple_messages := run_inputs.get("messages")):
        return
    assert isinstance(
        multiple_messages, Iterable
    ), f"expected Iterable, found {type(multiple_messages)}"
    # This will only get the first set of messages.
    if not (first_messages := next(iter(multiple_messages), None)):
        return
    assert isinstance(first_messages, Iterable), f"expected Iterable, found {type(first_messages)}"
    parsed_messages = []
    for message_data in first_messages:
        assert hasattr(message_data, "get"), f"expected Mapping, found {type(message_data)}"
        parsed_messages.append(_parse_message_data(message_data))
    if parsed_messages:
        yield LLM_INPUT_MESSAGES, parsed_messages


def _output_messages(run_outputs: Mapping[str, Any]) -> Iterator[Tuple[str, List[Message]]]:
    """Yields chat messages if present."""
    if not hasattr(run_outputs, "get"):
        return
    # There may be more than one set of generations. We'll use just the first set.
    if not (multiple_generations := run_outputs.get("generations")):
        return
    assert isinstance(
        multiple_generations, Iterable
    ), f"expected Iterable, found {type(multiple_generations)}"
    # This will only get the first set of generations.
    if not (first_generations := next(iter(multiple_generations), None)):
        return
    assert isinstance(
        first_generations, Iterable
    ), f"expected Iterable, found {type(first_generations)}"
    parsed_messages = []
    for generation in first_generations:
        assert hasattr(generation, "get"), f"expected Mapping, found {type(generation)}"
        if message_data := generation.get("message"):
            assert hasattr(message_data, "get"), f"expected Mapping, found {type(message_data)}"
            parsed_messages.append(_parse_message_data(message_data))
    if parsed_messages:
        yield LLM_OUTPUT_MESSAGES, parsed_messages


def _parse_message_data(message_data: Mapping[str, Any]) -> Message:
    """Parses message data to grab message role, content, etc."""
    message_class_name = message_data["id"][-1]
    if message_class_name.startswith("HumanMessage"):
        role = "user"
    elif message_class_name.startswith("AIMessage"):
        role = "assistant"
    elif message_class_name.startswith("SystemMessage"):
        role = "system"
    elif message_class_name.startswith("FunctionMessage"):
        role = "function"
    elif message_class_name.startswith("ChatMessage"):
        role = message_data["kwargs"]["role"]
    else:
        raise ValueError(f"Cannot parse message of type: {message_class_name}")
    parsed_message_data = {MESSAGE_ROLE: role}
    if kwargs := message_data.get("kwargs"):
        assert hasattr(kwargs, "get"), f"expected Mapping, found {type(kwargs)}"
        if content := kwargs.get("content"):
            assert isinstance(content, str), f"content must be str, found {type(content)}"
            parsed_message_data[MESSAGE_CONTENT] = content
        if additional_kwargs := kwargs.get("additional_kwargs"):
            assert hasattr(
                additional_kwargs, "get"
            ), f"expected Mapping, found {type(additional_kwargs)}"
            if function_call := additional_kwargs.get("function_call"):
                assert hasattr(
                    function_call, "get"
                ), f"expected Mapping, found {type(function_call)}"
                if name := function_call.get("name"):
                    assert isinstance(name, str), f"name must be str, found {type(name)}"
                    parsed_message_data[MESSAGE_FUNCTION_CALL_NAME] = name
                if arguments := function_call.get("arguments"):
                    assert isinstance(
                        arguments, str
                    ), f"arguments must be str, found {type(arguments)}"
                    parsed_message_data[MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON] = arguments
    return parsed_message_data


def _prompt_template(run_serialized: Dict[str, Any]) -> Iterator[Tuple[str, Any]]:
    """
    A best-effort attempt to locate the PromptTemplate object among the
    keyword arguments of a serialized object, e.g. an LLMChain object.
    """
    for obj in run_serialized.get("kwargs", {}).values():
        if not isinstance(obj, dict) or "id" not in obj:
            continue
        # The `id` field of the object is a list indicating the path to the
        # object's class in the LangChain package, e.g. `PromptTemplate` in
        # the `langchain.prompts.prompt` module is represented as
        # ["langchain", "prompts", "prompt", "PromptTemplate"]
        if obj["id"][-1].endswith("PromptTemplate"):
            kwargs = obj.get("kwargs", {})
            if not (template := kwargs.get("template", "")):
                continue
            yield LLM_PROMPT_TEMPLATE, template
            yield LLM_PROMPT_TEMPLATE_VARIABLES, kwargs.get("input_variables", [])
            yield LLM_PROMPT_TEMPLATE_VERSION, "unknown"
            break


def _invocation_parameters(run: Dict[str, Any]) -> Iterator[Tuple[str, str]]:
    """Yields invocation parameters if present."""
    if run["run_type"] != "llm":
        return
    run_extra = run["extra"]
    yield LLM_INVOCATION_PARAMETERS, json.dumps(run_extra.get("invocation_params", {}))


def _model_name(run_extra: Dict[str, Any]) -> Iterator[Tuple[str, str]]:
    """Yields model name if present."""
    if not (invocation_params := run_extra.get("invocation_params")):
        return
    for key in ["model_name", "model"]:
        if name := invocation_params.get(key):
            yield LLM_MODEL_NAME, name
            return


def _token_counts(run_outputs: Dict[str, Any]) -> Iterator[Tuple[str, int]]:
    """Yields token count information if present."""
    try:
        token_usage = run_outputs["llm_output"]["token_usage"]
    except Exception:
        return
    for attribute_name, key in [
        (LLM_TOKEN_COUNT_PROMPT, "prompt_tokens"),
        (LLM_TOKEN_COUNT_COMPLETION, "completion_tokens"),
        (LLM_TOKEN_COUNT_TOTAL, "total_tokens"),
    ]:
        if (token_count := token_usage.get(key)) is not None:
            yield attribute_name, token_count


def _function_calls(run_outputs: Dict[str, Any]) -> Iterator[Tuple[str, str]]:
    """Yields function call information if present."""
    try:
        function_call_data = deepcopy(
            run_outputs["generations"][0][0]["message"]["kwargs"]["additional_kwargs"][
                "function_call"
            ]
        )
        function_call_data["arguments"] = json.loads(function_call_data["arguments"])
        yield LLM_FUNCTION_CALL, json.dumps(function_call_data)
    except Exception:
        pass


def _tools(run: Dict[str, Any]) -> Iterator[Tuple[str, str]]:
    """Yields tool attributes if present."""
    if run["run_type"] != "tool":
        return
    run_serialized = run["serialized"]
    if "name" in run_serialized:
        yield TOOL_NAME, run_serialized["name"]
    if "description" in run_serialized:
        yield TOOL_DESCRIPTION, run_serialized["description"]
    # TODO: tool parameters https://github.com/Arize-ai/phoenix/issues/1330


def _retrieval_documents(
    run: Dict[str, Any],
) -> Iterator[Tuple[str, List[Any]]]:
    if run["run_type"] != "retriever":
        return
    yield (
        RETRIEVAL_DOCUMENTS,
        [
            {
                DOCUMENT_CONTENT: document.get("page_content"),
                DOCUMENT_METADATA: document.get("metadata") or {},
            }
            for document in (run.get("outputs") or {}).get("documents") or []
        ],
    )


def _chat_model_start_fallback(
    serialized: Dict[str, Any],
    messages: List[List[BaseMessage]],
    *,
    run_id: UUID,
    tags: Optional[List[str]] = None,
    parent_run_id: Optional[UUID] = None,
    metadata: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> None:
    # Currently does nothing. If a functional fallback is implemented, new failures will not be
    # caught
    pass


class OpenInferenceTracer(Tracer, BaseTracer):  # type: ignore
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._exporter = self._exporter or HttpExporter()

    def _convert_run_to_spans(
        self,
        run: Dict[str, Any],
        parent: Optional[Span] = None,
    ) -> None:
        attributes: Dict[str, Any] = {}
        for io_key, io_attributes in {
            "inputs": (INPUT_VALUE, INPUT_MIME_TYPE),
            "outputs": (OUTPUT_VALUE, OUTPUT_MIME_TYPE),
        }.items():
            attributes.update(zip(io_attributes, _convert_io(run.get(io_key))))
        attributes.update(_prompts(run["inputs"]))
        attributes.update(_input_messages(run["inputs"]))
        attributes.update(_output_messages(run["outputs"]))
        attributes.update(_prompt_template(run["serialized"]))
        attributes.update(_invocation_parameters(run))
        attributes.update(_model_name(run["extra"]))
        attributes.update(_token_counts(run["outputs"]))
        attributes.update(_function_calls(run["outputs"]))
        attributes.update(_tools(run))
        attributes.update(_retrieval_documents(run))
        events: List[SpanEvent] = []
        if (error := run["error"]) is None:
            status_code = SpanStatusCode.OK
        else:
            status_code = SpanStatusCode.ERROR
            # Since there is only one error message, keep just the
            # first error event.
            error_event = next(
                filter(
                    lambda event: event["name"] == "error",
                    run["events"],
                )
            )
            events.append(
                SpanException(
                    message=error,
                    timestamp=error_event["time"],
                )
            )
        span_kind = (
            SpanKind.AGENT
            if "agent" in run["name"].lower()
            else _langchain_run_type_to_span_kind(run["run_type"])
        )
        span = self.create_span(
            name=run["name"],
            span_kind=span_kind,
            parent_id=None if parent is None else parent.context.span_id,
            trace_id=None if parent is None else parent.context.trace_id,
            start_time=run["start_time"],
            end_time=run["end_time"],
            status_code=status_code,
            attributes=attributes,
            events=events,
        )
        for child_run in run["child_runs"]:
            self._convert_run_to_spans(child_run, span)

    def _persist_run(self, run: Run) -> None:
        # Note that this relies on `.dict()` from pydantic for the
        # serialization of objects like `langchain.schema.Document`.
        try:
            self._convert_run_to_spans(run.dict())
        except Exception:
            logger.exception("Failed to convert run to spans")

    @graceful_fallback(_chat_model_start_fallback)
    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        tags: Optional[List[str]] = None,
        parent_run_id: Optional[UUID] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Adds chat messages to the run inputs.

        LangChain's BaseTracer class does not implement hooks for chat models and hence does not
        record data such as the list of messages that were passed to the chat model.

        For reference, see https://github.com/langchain-ai/langchain/pull/4499.
        """

        parent_run_id_ = str(parent_run_id) if parent_run_id else None
        execution_order = self._get_execution_order(parent_run_id_)
        start_time = datetime.utcnow()
        if metadata:
            kwargs.update({"metadata": metadata})
        run = Run(
            id=run_id,
            parent_run_id=parent_run_id,
            serialized=serialized,
            inputs={"messages": [[dumpd(message) for message in batch] for batch in messages]},
            extra=kwargs,
            events=[{"name": "start", "time": start_time}],
            start_time=start_time,
            execution_order=execution_order,
            child_execution_order=execution_order,
            run_type="llm",
            tags=tags,
            name=name or "",
        )
        self._start_trace(run)



import weakref
from collections import defaultdict
from datetime import datetime, timezone
from enum import Enum
from queue import SimpleQueue
from threading import RLock, Thread
from types import MethodType
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    SupportsFloat,
    Tuple,
    Union,
    cast,
)
from uuid import UUID

from ddsketch import DDSketch
from sortedcontainers import SortedKeyList
from typing_extensions import TypeAlias
from wrapt import ObjectProxy

import phoenix.trace.v1 as pb
from phoenix.datetime_utils import right_open_time_range
from phoenix.trace import semantic_conventions
from phoenix.trace.schemas import (
    ATTRIBUTE_PREFIX,
    COMPUTED_PREFIX,
    CONTEXT_PREFIX,
    Span,
    SpanAttributes,
    SpanID,
    TraceID,
)
from phoenix.trace.v1.utils import decode, encode

END_OF_QUEUE = None  # sentinel value for queue termination

NAME = "name"
STATUS_CODE = "status_code"
SPAN_KIND = "span_kind"
TRACE_ID = CONTEXT_PREFIX + "trace_id"
SPAN_ID = CONTEXT_PREFIX + "span_id"
PARENT_ID = "parent_id"
START_TIME = "start_time"
END_TIME = "end_time"
LLM_TOKEN_COUNT_TOTAL = ATTRIBUTE_PREFIX + semantic_conventions.LLM_TOKEN_COUNT_TOTAL
LLM_TOKEN_COUNT_PROMPT = ATTRIBUTE_PREFIX + semantic_conventions.LLM_TOKEN_COUNT_PROMPT
LLM_TOKEN_COUNT_COMPLETION = ATTRIBUTE_PREFIX + semantic_conventions.LLM_TOKEN_COUNT_COMPLETION


class ComputedAttributes(Enum):
    # Enum value must be string prefixed by COMPUTED_PREFIX
    LATENCY_MS = (
        COMPUTED_PREFIX + "latency_ms"
    )  # The latency (or duration) of the span in milliseconds
    CUMULATIVE_LLM_TOKEN_COUNT_TOTAL = COMPUTED_PREFIX + "cumulative_token_count.total"
    CUMULATIVE_LLM_TOKEN_COUNT_PROMPT = COMPUTED_PREFIX + "cumulative_token_count.prompt"
    CUMULATIVE_LLM_TOKEN_COUNT_COMPLETION = COMPUTED_PREFIX + "cumulative_token_count.completion"


class ReadableSpan(ObjectProxy):  # type: ignore
    """
    A wrapped a protobuf Span, with access methods and ability to decode to
    a python span. It's meant to be interface layer separating use from
    implementation. It can also provide computed values that are not intrinsic
    to the span, e.g. the latency rank percent which can change as more spans
    are ingested, and would need to be re-computed on the fly.
    """

    __wrapped__: pb.Span

    def __init__(self, span: pb.Span) -> None:
        super().__init__(span)
        self._self_computed_values: Dict[str, SupportsFloat] = {}

    @property
    def span(self) -> Span:
        span = decode(self.__wrapped__)
        span.attributes.update(cast(SpanAttributes, self._self_computed_values))
        # TODO: compute latency rank percent (which can change depending on how
        # many spans already ingested).
        return span

    def __getitem__(self, key: str) -> Any:
        if key.startswith(COMPUTED_PREFIX):
            return self._self_computed_values.get(key)
        if key.startswith(CONTEXT_PREFIX):
            suffix_key = key[len(CONTEXT_PREFIX) :]
            return getattr(self.__wrapped__.context, suffix_key, None)
        if key.startswith(ATTRIBUTE_PREFIX):
            suffix_key = key[len(ATTRIBUTE_PREFIX) :]
            if suffix_key not in self.__wrapped__.attributes:
                return None
            return self.__wrapped__.attributes[suffix_key]
        return getattr(self.__wrapped__, key, None)

    def __setitem__(self, key: str, value: Any) -> None:
        if not key.startswith(COMPUTED_PREFIX):
            raise KeyError(f"{key} is not a computed value")
        self._self_computed_values[key] = value


ParentSpanID: TypeAlias = SpanID
ChildSpanID: TypeAlias = SpanID


class Traces:
    def __init__(self, spans: Optional[Iterable[Span]] = None) -> None:
        self._queue: "SimpleQueue[Optional[pb.Span]]" = SimpleQueue()
        # Putting `None` as the sentinel value for queue termination.
        weakref.finalize(self, self._queue.put, END_OF_QUEUE)
        for span in spans or ():
            self.put(span)
        self._lock = RLock()
        self._spans: Dict[SpanID, ReadableSpan] = {}
        self._parent_span_ids: Dict[SpanID, ParentSpanID] = {}
        self._traces: Dict[TraceID, List[SpanID]] = defaultdict(list)
        self._child_span_ids: DefaultDict[SpanID, List[ChildSpanID]] = defaultdict(list)
        self._orphan_spans: DefaultDict[ParentSpanID, List[pb.Span]] = defaultdict(list)
        self._start_time_sorted_span_ids: SortedKeyList[SpanID] = SortedKeyList(
            key=lambda span_id: self._spans[span_id].start_time.ToDatetime(timezone.utc),
        )
        self._start_time_sorted_root_span_ids: SortedKeyList[SpanID] = SortedKeyList(
            key=lambda span_id: self._spans[span_id].start_time.ToDatetime(timezone.utc),
        )
        self._latency_sorted_root_span_ids: SortedKeyList[SpanID] = SortedKeyList(
            key=lambda span_id: self._spans[span_id][ComputedAttributes.LATENCY_MS.value],
        )
        self._root_span_latency_ms_sketch = DDSketch()
        self._min_start_time: Optional[datetime] = None
        self._max_start_time: Optional[datetime] = None
        self._start_consumer()

    def put(self, span: Optional[Union[Span, pb.Span]] = None) -> None:
        self._queue.put(encode(span) if isinstance(span, Span) else span)

    def get_trace(self, trace_id: TraceID) -> Iterator[Span]:
        for span_id in self._traces[trace_id]:
            if span := self[span_id]:
                yield span

    def get_spans(
        self,
        start_time: Optional[datetime] = None,
        stop_time: Optional[datetime] = None,
        root_spans_only: Optional[bool] = False,
    ) -> Iterator[Span]:
        if not self._spans:
            return
        min_start_time, max_stop_time = cast(
            Tuple[datetime, datetime],
            self.right_open_time_range,
        )
        start_time = start_time or min_start_time
        stop_time = stop_time or max_stop_time
        sorted_span_ids = (
            self._start_time_sorted_root_span_ids
            if root_spans_only
            else self._start_time_sorted_span_ids
        )
        for span_id in sorted_span_ids.irange_key(
            start_time.astimezone(timezone.utc),
            stop_time.astimezone(timezone.utc),
            inclusive=(True, False),
            reverse=True,  # most recent spans first
        ):
            if span := self[span_id]:
                yield span

    def latency_rank_percent(self, latency_ms: float) -> Optional[float]:
        """
        Returns a value between 0 and 100 approximating the rank of the
        latency value as percent of the total count of root spans. E.g., for
        a latency value at the 75th percentile, the result is roughly 75.
        """
        root_span_ids = self._latency_sorted_root_span_ids
        if not (n := len(root_span_ids)):
            return None
        rank = cast(int, root_span_ids.bisect_key_left(latency_ms))
        return rank / n * 100

    def root_span_latency_ms_quantiles(self, *probabilities: float) -> Iterator[Optional[float]]:
        """Root span latency quantiles in milliseconds"""
        return map(self._root_span_latency_ms_sketch.get_quantile_value, probabilities)

    def get_descendant_span_ids(self, span_id: SpanID) -> Iterator[SpanID]:
        for child_span_id in self._child_span_ids.get(span_id) or ():
            yield child_span_id
            yield from self.get_descendant_span_ids(child_span_id)

    @property
    def span_count(self) -> int:
        """Total number of spans (excluding orphan spans if any)"""
        return len(self._spans)

    @property
    def token_count_total(self) -> int:
        count = 0
        for span in self._spans.values():
            count += span[LLM_TOKEN_COUNT_TOTAL] or 0
        return count

    @property
    def right_open_time_range(self) -> Tuple[Optional[datetime], Optional[datetime]]:
        return right_open_time_range(self._min_start_time, self._max_start_time)

    def __getitem__(self, span_id: SpanID) -> Optional[Span]:
        with self._lock:
            if span := self._spans.get(span_id):
                return span.span
        return None

    def _start_consumer(self) -> None:
        Thread(
            target=MethodType(
                self.__class__._consume_spans,
                weakref.proxy(self),
            ),
            daemon=True,
        ).start()

    def _consume_spans(self) -> None:
        while (item := self._queue.get()) is not END_OF_QUEUE:
            with self._lock:
                self._process_span(item)

    def _process_span(self, span: pb.Span) -> None:
        span_id = UUID(bytes=span.context.span_id)
        existing_span = self._spans.get(span_id)
        if existing_span and existing_span.end_time:
            # Reject updates if span has ended.
            return
        is_root_span = not span.HasField("parent_span_id")
        if not is_root_span:
            parent_span_id = UUID(bytes=span.parent_span_id.value)
            if parent_span_id not in self._spans:
                # Span can't be processed before its parent.
                self._orphan_spans[parent_span_id].append(span)
                return
            self._child_span_ids[parent_span_id].append(span_id)
            self._parent_span_ids[span_id] = parent_span_id
        new_span = ReadableSpan(span)
        start_time = span.start_time.ToDatetime(timezone.utc)
        end_time = span.end_time.ToDatetime(timezone.utc) if span.HasField("end_time") else None
        if end_time:
            new_span[ComputedAttributes.LATENCY_MS.value] = latency = (
                end_time - start_time
            ).total_seconds() * 1000
            if is_root_span:
                self._root_span_latency_ms_sketch.add(latency)
        self._spans[span_id] = new_span
        if is_root_span and end_time:
            self._latency_sorted_root_span_ids.add(span_id)
        if not existing_span:
            trace_id = UUID(bytes=span.context.trace_id)
            self._traces[trace_id].append(span_id)
            if is_root_span:
                self._start_time_sorted_root_span_ids.add(span_id)
            self._start_time_sorted_span_ids.add(span_id)
            self._min_start_time = (
                start_time
                if self._min_start_time is None
                else min(self._min_start_time, start_time)
            )
            self._max_start_time = (
                start_time
                if self._max_start_time is None
                else max(self._max_start_time, start_time)
            )
        # Update cumulative values for span's ancestors.
        for attribute_name, cumulative_attribute_name in (
            (LLM_TOKEN_COUNT_TOTAL, ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_TOTAL.value),
            (LLM_TOKEN_COUNT_PROMPT, ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_PROMPT.value),
            (
                LLM_TOKEN_COUNT_COMPLETION,
                ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_COMPLETION.value,
            ),
        ):
            existing_value = (existing_span[attribute_name] or 0) if existing_span else 0
            new_value = new_span[attribute_name] or 0
            if not (difference := new_value - existing_value):
                continue
            existing_cumulative_value = (
                (existing_span[cumulative_attribute_name] or 0) if existing_span else 0
            )
            new_span[cumulative_attribute_name] = difference + existing_cumulative_value
            self._add_value_to_span_ancestors(
                span_id,
                cumulative_attribute_name,
                difference,
            )
        # Process previously orphaned spans, if any.
        for orphan_span in self._orphan_spans[span_id]:
            self._process_span(orphan_span)

    def _add_value_to_span_ancestors(
        self,
        span_id: SpanID,
        attribute_name: str,
        value: float,
    ) -> None:
        while parent_span_id := self._parent_span_ids.get(span_id):
            parent_span = self._spans[parent_span_id]
            cumulative_value = parent_span[attribute_name] or 0
            parent_span[attribute_name] = cumulative_value + value
            span_id = parent_span_id



import ast
from typing import Any, Iterator, Mapping, Tuple, cast

from phoenix.core.traces import ComputedAttributes
from phoenix.trace import semantic_conventions
from phoenix.trace.schemas import COMPUTED_PREFIX, Span


class SpanFilter:
    def __init__(self, condition: str) -> None:
        self._root = ast.parse(condition, mode="eval")
        _validate_expression(self._root, condition)
        self._translated = _Translator(condition).visit(self._root)
        ast.fix_missing_locations(self._translated)
        self._compiled = compile(self._translated, filename="", mode="eval")

    def __call__(self, span: Span) -> bool:
        return cast(bool, eval(self._compiled, {}, {"span": span, "_MISSING": _Missing()}))


def _replace_none_with_missing(
    value: ast.expr,
    as_str: bool = False,
) -> ast.IfExp:
    """
    E.g. `value` becomes
    `_MISSING if (_MAYBE := value) is None else _MAYBE`
    """
    _store_MAYBE = ast.Name(id="_MAYBE", ctx=ast.Store())
    _load_MAYBE = ast.Name(id="_MAYBE", ctx=ast.Load())
    return ast.IfExp(
        test=ast.Compare(
            left=ast.NamedExpr(target=_store_MAYBE, value=value),
            ops=[ast.Is()],
            comparators=[ast.Constant(value=None)],
        ),
        body=ast.Name(id="_MISSING", ctx=ast.Load()),
        orelse=_as_str(_load_MAYBE) if as_str else _load_MAYBE,
    )


def _as_str(value: ast.expr) -> ast.Call:
    """E.g. `value` becomes `str(value)`"""
    return ast.Call(func=ast.Name(id="str", ctx=ast.Load()), args=[value], keywords=[])


def _ast_replacement(expression: str) -> ast.expr:
    as_str = expression in (
        "span.status_code",
        "span.span_kind",
        "span.parent_id",
        "span.context.span_id",
        "span.context.trace_id",
    )
    return _replace_none_with_missing(ast.parse(expression, mode="eval").body, as_str)


def _allowed_replacements() -> Iterator[Tuple[str, ast.expr]]:
    for source_segment, ast_replacement in {
        "name": _ast_replacement("span.name"),
        "status_code": _ast_replacement("span.status_code"),
        "span_kind": _ast_replacement("span.span_kind"),
        "parent_id": _ast_replacement("span.parent_id"),
    }.items():
        yield source_segment, ast_replacement
        yield "span." + source_segment, ast_replacement

    for source_segment, ast_replacement in {
        "span_id": _ast_replacement("span.context.span_id"),
        "trace_id": _ast_replacement("span.context.trace_id"),
    }.items():
        yield source_segment, ast_replacement
        yield "context." + source_segment, ast_replacement
        yield "span.context." + source_segment, ast_replacement

    for field_name in (
        getattr(semantic_conventions, variable_name)
        for variable_name in dir(semantic_conventions)
        if variable_name.isupper()
    ):
        source_segment = field_name
        ast_replacement = _ast_replacement(f"span.attributes.get('{field_name}')")
        yield source_segment, ast_replacement
        yield "attributes." + source_segment, ast_replacement
        yield "span.attributes." + source_segment, ast_replacement

    for computed_attribute in ComputedAttributes:
        field_name = computed_attribute.value
        source_segment = field_name[len(COMPUTED_PREFIX) :]
        ast_replacement = _ast_replacement(f"span.attributes.get('{field_name}')")
        yield source_segment, ast_replacement


class _Translator(ast.NodeTransformer):
    _allowed_fields: Mapping[str, ast.expr] = dict(_allowed_replacements())

    def __init__(self, source: str) -> None:
        # In Python 3.8, we have to use `ast.get_source_segment(source, node)`.
        # In Python 3.9, we can use `ast.unparse(node)` instead.
        self._source = source

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        source_segment: str = cast(str, ast.get_source_segment(self._source, node))
        if replacement := self._allowed_fields.get(source_segment):
            return replacement
        raise SyntaxError(f"invalid expression: {source_segment}")  # TODO: add details

    def visit_Name(self, node: ast.Name) -> Any:
        source_segment: str = cast(str, ast.get_source_segment(self._source, node))
        if replacement := self._allowed_fields.get(source_segment):
            return replacement
        raise SyntaxError(f"invalid expression: {source_segment}")  # TODO: add details

    def visit_Constant(self, node: ast.Constant) -> Any:
        return ast.Name(id="_MISSING", ctx=ast.Load()) if node.value is None else node


def _validate_expression(expression: ast.Expression, source: str) -> None:
    # In Python 3.8, we have to use `ast.get_source_segment(source, node)`.
    # In Python 3.9, we can use `ast.unparse(node)` instead.
    if not isinstance(expression, ast.Expression):
        raise SyntaxError(f"invalid expression: {source}")  # TODO: add details
    for i, node in enumerate(ast.walk(expression.body)):
        if i == 0:
            if isinstance(node, (ast.BoolOp, ast.Compare)):
                continue
        elif isinstance(
            node,
            (
                ast.Attribute,
                ast.BinOp,
                ast.BoolOp,
                ast.Compare,
                ast.Constant,
                ast.Load,
                ast.Name,
                ast.Tuple,
                ast.List,
                ast.UnaryOp,
                ast.boolop,
                ast.cmpop,
                ast.operator,
                ast.unaryop,
            ),
        ):
            continue
        source_segment = cast(str, ast.get_source_segment(source, node))
        raise SyntaxError(f"invalid expression: {source_segment}")  # TODO: add details


class _Missing:
    """Falsifies all comparisons except those with self."""

    def __lt__(self, other: Any) -> bool:
        return False

    def __le__(self, other: Any) -> bool:
        return False

    def __gt__(self, other: Any) -> bool:
        return False

    def __ge__(self, other: Any) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, _Missing)

    def __ne__(self, other: Any) -> bool:
        return False

    def __len__(self) -> int:
        return 0

    def __iter__(self) -> Any:
        return self

    def __next__(self) -> Any:
        raise StopIteration()

    def __contains__(self, item: Any) -> bool:
        return False

    def __str__(self) -> str:
        return ""

    def __float__(self) -> float:
        return float("nan")



import gzip
import logging
import weakref
from queue import SimpleQueue
from threading import Thread
from types import MethodType
from typing import Any, Optional

import requests
from requests import Session

import phoenix.trace.v1 as pb
from phoenix.config import get_env_host, get_env_port
from phoenix.trace.schemas import Span
from phoenix.trace.v1.utils import encode

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

END_OF_QUEUE = None  # sentinel value for queue termination


class NoOpExporter:
    def export(self, _: Any) -> None:
        pass


class HttpExporter:
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> None:
        """
        Span Exporter using HTTP.

        Parameters
        ----------
        host: Optional[str]
            The host of the Phoenix server. It can also be set using environment
            variable `PHOENIX_HOST`, otherwise it defaults to `127.0.0.1`.
        port: Optional[int]
            The port of the Phoenix server. It can also be set using environment
            variable `PHOENIX_PORT`, otherwise it defaults to `6006`.
        """
        self._host = host or get_env_host()
        self._port = port or get_env_port()
        self._base_url = f"http://{self._host}:{self._port}"
        self._warn_if_phoenix_is_not_running()
        self._session = Session()
        weakref.finalize(self, self._session.close)
        self._session.headers.update(
            {
                "content-type": "application/x-protobuf",
                "content-encoding": "gzip",
            }
        )
        self._queue: "SimpleQueue[Optional[pb.Span]]" = SimpleQueue()
        # Putting `None` as the sentinel value for queue termination.
        weakref.finalize(self, self._queue.put, END_OF_QUEUE)
        self._start_consumer()

    def export(self, span: Span) -> None:
        self._queue.put(encode(span))

    def _start_consumer(self) -> None:
        Thread(
            target=MethodType(
                self.__class__._consume_items,
                weakref.proxy(self),
            ),
            daemon=True,
        ).start()

    def _consume_items(self) -> None:
        while (item := self._queue.get()) is not END_OF_QUEUE:
            self._send(item)

    def _send(self, item: pb.Span) -> None:
        serialized = item.SerializeToString()
        data = gzip.compress(serialized)
        try:
            self._session.post(self._url(item), data=data)
        except Exception as e:
            logger.exception(e)

    def _url(self, _: pb.Span) -> str:
        return f"{self._base_url}/v1/spans"

    def _warn_if_phoenix_is_not_running(self) -> None:
        try:
            requests.get(f"{self._base_url}/arize_phoenix_version").raise_for_status()
        except Exception:
            logger.warning(
                f"Arize Phoenix is not running on {self._base_url}. Launch Phoenix "
                f"with `import phoenix as px; px.launch_app()`"
            )
